"""Tests that Roles hosted by Pulp can be installed by ansible-galaxy."""

import pytest
from os import path
import subprocess

from pulp_ansible.tests.functional.constants import (
    ANSIBLE_ELASTIC_FIXTURE_URL,
    ANSIBLE_ELASTIC_ROLE_NAMESPACE_NAME,
    ANSIBLE_ELASTIC_ROLE,
)


@pytest.mark.parallel
def test_install_role(
    ansible_distribution_factory,
    ansible_role_remote_factory,
    ansible_sync_factory,
    ansible_dir_factory,
):
    """Test whether ansible-galaxy can install a Role hosted by Pulp."""
    distribution = ansible_distribution_factory(
        repository=ansible_sync_factory(
            remote=ansible_role_remote_factory(url=ANSIBLE_ELASTIC_FIXTURE_URL).pulp_href
        )
    )

    ansible_dir = ansible_dir_factory(distribution.client_url)
    directory = ansible_dir / ANSIBLE_ELASTIC_ROLE_NAMESPACE_NAME

    assert not path.exists(directory), "Directory {} already exists".format(directory)
    subprocess.run(
        (
            "ansible-galaxy",
            "role",
            "install",
            ANSIBLE_ELASTIC_ROLE,
            "-c",
            "-p",
            ansible_dir,
        ),
        check=True,
    )

    assert path.exists(directory), "Could not find directory {}".format(directory)



"""Tests related to upload of collections."""

import hashlib
import pathlib
import re
import time
from datetime import datetime
from urllib.parse import urljoin

import pytest
import requests
from django.conf import settings

DOMAIN_ENABLED = settings.DOMAIN_ENABLED


class NullAuth(requests.auth.AuthBase):
    def __call__(self, r):
        return r


class BaseURLSession(requests.Session):
    def __init__(self, base_url, *args, **kwargs):
        self.base_url = base_url
        super().__init__(*args, **kwargs)

    def request(self, method, url, **kwargs):
        return super().request(method, urljoin(self.base_url, url), **kwargs)


@pytest.fixture(scope="module")
def http_session(bindings_cfg):
    with BaseURLSession(bindings_cfg.host) as session:
        session.auth = (bindings_cfg.username, bindings_cfg.password)
        yield session


def get_galaxy_url(base, path, domain="default"):
    """Given an Ansible Distribution base_path and an endpoint path, construct a URL.

    Takes the expected GALAXY_API_ROOT setting of the target into consideration.
    """
    GALAXY_API_ROOT = (
        f"/pulp_ansible/galaxy/{domain}/{base}/api/"
        if DOMAIN_ENABLED
        else f"/pulp_ansible/galaxy/{base}/api/"
    )
    return urljoin(GALAXY_API_ROOT, path.lstrip("/"))


@pytest.fixture(scope="module")
def collection_artifact(ansible_collection_factory):
    """Generate a randomized collection for testing."""
    return ansible_collection_factory()


@pytest.fixture(scope="module")
def collection_artifact2(ansible_collection_factory):
    """
    Generate a second randomized collection for testing.

    This collection will have the same namespace and version, but different name.
    """
    return ansible_collection_factory()


def get_metadata_published(http_session, pulp_dist):
    """Return published datetime."""
    response = http_session.get(get_galaxy_url(pulp_dist.base_path, "/v3/"))
    response.raise_for_status()
    metadata = response.json()
    return datetime.strptime(metadata["published"], "%Y-%m-%dT%H:%M:%S.%fZ")


def upload_collection(http_session, filename, base_path):
    """Helper to upload collections to pulp_ansible/galaxy."""
    UPLOAD_PATH = get_galaxy_url(base_path, "/v3/artifacts/collections/")
    with open(filename, "rb") as fp:
        collection = {"file": fp}
        response = http_session.post(UPLOAD_PATH, files=collection)
    response.raise_for_status()
    task = response.json()["task"]
    while True:
        response = http_session.get(task)
        response.raise_for_status()
        result = response.json()
        if result["state"] == "running":
            time.sleep(1)
            continue
        if result["state"] == "completed":
            break
        raise Exception(str(result))
    return result


@pytest.fixture(scope="class")
def collection_upload(http_session, collection_artifact, pulp_dist):
    """Publish a new collection and return the processed response data."""
    published_before_upload = get_metadata_published(http_session, pulp_dist)
    response = upload_collection(http_session, collection_artifact.filename, pulp_dist.base_path)
    published_after_upload = get_metadata_published(http_session, pulp_dist)
    assert published_after_upload > published_before_upload
    return response


@pytest.fixture(scope="class")
def collection_upload2(http_session, collection_artifact2, pulp_dist):
    """Publish the second new collection and return the processed response data."""
    published_before_upload = get_metadata_published(http_session, pulp_dist)
    response = upload_collection(http_session, collection_artifact2.filename, pulp_dist.base_path)
    published_after_upload = get_metadata_published(http_session, pulp_dist)
    assert published_after_upload > published_before_upload
    return response


@pytest.fixture(scope="class")
def collection_detail(http_session, collection_upload, pulp_dist, collection_artifact):
    """Fetch and parse a collection details response from an uploaded collection."""
    url = get_galaxy_url(
        pulp_dist.base_path,
        f"/v3/collections/{collection_artifact.namespace}/{collection_artifact.name}/",
    )
    response = http_session.get(url)
    response.raise_for_status()
    return response.json()


@pytest.fixture(scope="class")
def collection_highest_version(http_session, collection_detail):
    response = http_session.get(collection_detail["highest_version"]["href"])
    response.raise_for_status()
    return response.json()


@pytest.fixture(scope="class")
def pulp_dist(ansible_repository_factory, ansible_distribution_factory):
    """Create an Ansible Distribution to simulate the automation hub environment for testing."""
    return ansible_distribution_factory(repository=ansible_repository_factory())


class TestCollection:

    def test_collection_upload(self, collection_upload):
        """Upload a new collection.

        Uploads a newly generated collection and validates the resulting collection version details.
        """
        # Validate the upload response
        assert collection_upload["error"] is None
        assert re.match(
            r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}", collection_upload["id"]
        )
        assert collection_upload["state"] == "completed"

        assert "updated_at" in collection_upload
        assert "started_at" in collection_upload
        assert "created_at" in collection_upload
        assert "finished_at" in collection_upload
        assert "messages" in collection_upload

        # TODO: Add this back when namespace, name, and version are apart of the CollectionImport
        # for key, value in collection_upload.items():
        #     if key in COLLECTION_METADATA.keys():
        #         assert COLLECTION_METADATA[key] == value, collection_upload

    def test_collection_list(
        self,
        http_session,
        collection_artifact,
        collection_artifact2,
        collection_upload,
        collection_upload2,
        pulp_dist,
    ):
        """Tests the collection list endpoint after uploading both collections."""
        url = get_galaxy_url(pulp_dist.base_path, "v3/collections/")
        response = http_session.get(url)
        response.raise_for_status()
        result = response.json()

        assert result["meta"]["count"] >= 2
        present_collections = {c["href"].split("collections/")[1] for c in result["data"]}
        uploaded_collections = {
            f"index/{collection_artifact.namespace}/{collection_artifact.name}/",
            f"index/{collection_artifact2.namespace}/{collection_artifact2.name}/",
        }
        assert uploaded_collections.issubset(present_collections)

    def test_collection_detail(self, collection_artifact, collection_detail, pulp_dist):
        """Test collection detail resulting from a successful upload of one version.

        Includes information of the most current version.
        """
        url = (
            f"plugin/ansible/content/{pulp_dist.base_path}"
            f"/collections/index/{collection_artifact.namespace}/{collection_artifact.name}/"
        )

        assert not collection_detail["deprecated"]

        # Check that the URL ends with the correct path so that this test doesn't fail
        # when galaxy_ng is installed
        assert collection_detail["href"].endswith(url)
        assert collection_detail["namespace"] == collection_artifact.namespace
        assert collection_detail["name"] == collection_artifact.name
        assert collection_detail["highest_version"]["version"] == "1.0.0"
        assert collection_detail["download_count"] == 0

    def test_collection_version_list(
        self, http_session, collection_artifact, collection_detail, collection_upload2
    ):
        """Test the versions endpoint, listing the available versions of a given collection."""
        # Version List Endpoint
        response = http_session.get(collection_detail["versions_url"])
        response.raise_for_status()
        versions = response.json()
        assert versions["meta"]["count"] == 1
        version = versions["data"][0]

        assert version["version"] == "1.0.0"
        assert version["href"] == collection_detail["highest_version"]["href"]

    def test_collection_version_filter_by_q(
        self,
        ansible_bindings,
        ansible_collection_factory,
        pulp_dist,
        monitor_task,
        delete_orphans_pre,
    ):
        """Verify successive imports do not aggregate tags into search vectors."""

        def publish(new_artifact):
            body = {
                "file": new_artifact.filename,
                "expected_namespace": new_artifact.namespace,
                "expected_name": new_artifact.name,
                "expected_version": new_artifact.version,
            }
            resp = ansible_bindings.ContentCollectionVersionsApi.create(**body)
            monitor_task(resp.task)

        # make&publish 2 collections with each having unique tags
        specs = [("tag1", "ns1", "col1"), ("tag2", "ns1", "col2")]
        for spec in specs:
            cfg = {
                "namespace": spec[1],
                "name": spec[2],
                "description": "",
                "repository": f"https://github.com/{spec[1]}/{spec[2]}",
                "authors": ["jimbob"],
                "version": "1.0.0",
                "tags": [spec[0]],
            }
            this_artifact = ansible_collection_factory(config=cfg)
            publish(this_artifact)

        for spec in specs:
            resp = ansible_bindings.ContentCollectionVersionsApi.list(q=spec[0])

            # should only get the 1 cv as a result ...
            assert resp.count == 1
            assert resp.results[0].namespace == spec[1]
            assert resp.results[0].name == spec[2]

    def test_collection_version(self, http_session, collection_artifact, collection_detail):
        """Test collection version endpoint.

        Each collection version details a specific uploaded artifact for the collection.
        """
        # Version Endpoint
        response = http_session.get(collection_detail["highest_version"]["href"])
        response.raise_for_status()
        version = response.json()

        assert version["name"] == collection_artifact.name
        assert version["namespace"] == {
            "metadata_sha256": None,
            "name": collection_artifact.namespace,
        }
        assert version["version"] == "1.0.0"
        with open(collection_artifact.filename, "rb") as fp:
            tarball = fp.read()
        assert version["artifact"]["sha256"] == hashlib.sha256(tarball).hexdigest()
        assert version["artifact"]["size"] == len(tarball)

        assert version["artifact"]["filename"] == pathlib.Path(collection_artifact.filename).name

        assert "updated_at" in version
        assert "created_at" in version

        assert "files" in version
        assert "manifest" in version
        assert "requires_ansible" in version

        #     # TODO: Test meta data
        #     # 'metadata': {'authors': ['Orion User 1'],
        #     #             'contents': [],
        #     #             'dependencies': {},
        #     #             'description': 'a collection with some deps on other collections',
        #     #             'documentation': '',
        #     #             'homepage': '',
        #     #             'issues': '',
        #     #             'license': ['GPL-3.0-or-later'],
        #     #             'repository': 'http://github.example.com/orionuser1/skeleton',
        #     #             'tags': ['collectiontest']},

    def test_collection_download_metadata_unauthorized_fails(
        self,
        http_session,
        collection_detail,
    ):
        response = http_session.get(collection_detail["highest_version"]["href"], auth=NullAuth())
        assert response.status_code == 401

    def test_collection_download_redirects(
        self,
        http_session,
        collection_highest_version,
        collection_artifact,
    ):
        # Artifact Download Endoint
        url = collection_highest_version["download_url"]

        with open(collection_artifact.filename, "rb") as fp:
            tarball = fp.read()

        response = http_session.get(url)
        assert response.status_code == 200, (url, response.request.headers)
        assert response.content == tarball

    def test_collection_download_redirect_fails_unauthorized(
        self,
        http_session,
        collection_highest_version,
    ):
        # Artifact Download Endoint
        url = collection_highest_version["download_url"]

        response = http_session.get(url, auth=NullAuth())
        assert response.status_code == 401

    def test_downloading_collection_with_bad_name_returns_not_found(
        self,
        http_session,
        collection_highest_version,
    ):
        # Artifact Download Endoint, but we mess up the address.
        url = collection_highest_version["download_url"].replace("-", "_")

        response = http_session.get(url)
        assert response.status_code == 404

    def test_collection_upload_repeat(
        self, http_session, ansible_collection_factory, pulp_dist, collection_upload
    ):
        """
        Upload a duplicate collection.
        """
        response = upload_collection(
            http_session, ansible_collection_factory().filename, pulp_dist.base_path
        )
        assert response["error"] is None
        assert response["state"] == "completed"
        assert re.match(
            r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}", response["id"]
        )



from django.test import TestCase

from pulp_ansible.app.models import AnsibleDistribution
from pulp_ansible.app.models import AnsibleRepository
from pulp_ansible.app.models import CollectionVersion
from pulp_ansible.app.models import CrossRepositoryCollectionVersionIndex as CVIndex

from .utils import build_cvs_from_specs, randstr


class TestCollectionVersionIndex(TestCase):
    """Test Collection Version Cross Repository Index Behavior."""

    def test_cv_index_removes_cvs(self):
        """Ensure when CVs are removed from a repo, the index is also removed."""
        repo_name = randstr()
        base_path = repo_name
        distro_name = repo_name
        ns = randstr()

        # create a repository
        repo = AnsibleRepository(name=repo_name)
        repo.save()

        # create the distro
        AnsibleDistribution.objects.create(name=distro_name, base_path=base_path, repository=repo)

        # add a collection version ...
        specs = [
            (ns, "foo", "1.0.0"),
            (ns, "bar", "1.0.0"),
        ]
        collection_versions = build_cvs_from_specs(specs)
        cv_pks = [x.pk for x in collection_versions]
        qs = CollectionVersion.objects.filter(pk__in=cv_pks)
        with repo.new_version() as new_version:
            new_version.add_content(qs)

        # ensure CVs were indexed
        assert CVIndex.objects.filter(repository=repo, repository_version=None).count() == 2

        # remove foo
        to_remove = CollectionVersion.objects.filter(namespace=ns, name="foo", version="1.0.0")
        with repo.new_version() as new_version:
            new_version.remove_content(to_remove)

        # ensure only one index is left for bar
        qs = CVIndex.objects.filter(repository=repo, repository_version=None).order_by(
            "collection_version__version"
        )
        assert qs.count() == 1
        assert qs.first().collection_version.namespace == ns
        assert qs.first().collection_version.name == "bar"
        assert qs.first().collection_version.version == "1.0.0"

    def test_cv_index_retains_ids(self):
        """Ensure old indexes are retained instead of deleted and recreated."""
        repo_name = randstr()
        base_path = repo_name
        distro_name = repo_name
        ns = randstr()

        # create a repository
        repo = AnsibleRepository(name=repo_name)
        repo.save()

        # create the distro
        AnsibleDistribution.objects.create(name=distro_name, base_path=base_path, repository=repo)

        # add a collection version ...
        specs = [
            (ns, "bar", "1.0.0"),
        ]
        collection_versions = build_cvs_from_specs(specs)
        cv_pks = [x.pk for x in collection_versions]
        qs = CollectionVersion.objects.filter(pk__in=cv_pks)
        with repo.new_version() as new_version:
            new_version.add_content(qs)

        # ensure CVs were indexed
        qs = CVIndex.objects.filter(repository=repo, repository_version=None)
        assert qs.count() == 1

        old_ids = [x.id for x in qs]

        # add 1 more collection version ...
        specs = [
            (ns, "bar", "2.0.0"),
        ]
        collection_versions = build_cvs_from_specs(specs)
        cv_pks = [x.pk for x in collection_versions]
        qs = CollectionVersion.objects.filter(pk__in=cv_pks)
        with repo.new_version() as new_version:
            new_version.add_content(qs)

        # get the indexes and their IDs again ...
        qs = CVIndex.objects.filter(repository=repo, repository_version=None).order_by(
            "collection_version__version"
        )
        assert qs.count() == 2

        new_indexes = [x for x in qs]
        new_ids = [x.id for x in qs]

        # the previous IDs should be in the list ...
        for x in old_ids:
            assert x in new_ids

        assert new_indexes[0].is_highest is False
        assert new_indexes[0].collection_version.version == "1.0.0"
        assert new_indexes[1].is_highest is True
        assert new_indexes[1].collection_version.version == "2.0.0"



import pytest


@pytest.mark.parallel
def test_client_config_distro(ansible_bindings, ansible_repo_factory, ansible_distribution_factory):
    distribution = ansible_distribution_factory(ansible_repo_factory())
    response = ansible_bindings.PulpAnsibleApiV3PluginAnsibleClientConfigurationApi.read(
        path=distribution.base_path
    )
    assert response.default_distribution_path == distribution.base_path


@pytest.mark.parallel
def test_client_config_default_distro(ansible_bindings):
    response = ansible_bindings.PulpAnsibleDefaultApiV3PluginAnsibleClientConfigurationApi.read()
    assert response.default_distribution_path is None



import argparse
import os

import django

django.setup()

from pulpcore.plugin.tasking import (  # noqa otherwise E402: module level not at top of file
    dispatch,
)

from pulp_ansible.app.tasks.test_tasks import (  # noqa otherwise E402: module level not at top of file
    import_collection_from_path,
)


parser = argparse.ArgumentParser(description="Quickly load collections form a folder.")

parser.add_argument(
    "--collections-dir",
    metavar="PATH",
    type=str,
    nargs=1,
    required=True,
    help="The full path to a director containing collection tarballs.",
)

args = parser.parse_args()


if __name__ == "__main__":
    for root, dirs, files in os.walk(args.collections_dir[0]):
        for file in files:
            if file.endswith(".tar.gz"):
                async_result = dispatch(
                    import_collection_from_path, args=(os.path.join(root, file),)
                )



# Generated by Django 2.2.4 on 2019-08-02 22:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0005_collectionversion_is_highest'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversion',
            name='name',
            field=models.CharField(editable=False, max_length=64),
        ),
        migrations.RemoveField(
            model_name='collectionremote',
            name='whitelist',
        ),
    ]



# Generated by Django 4.2.22 on 2025-07-09 10:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0064_alter_collection_unique_together_and_more"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="collectionversion",
            index=models.Index(
                fields=["namespace", "name", "version"], name="ansible_col_namespa_ea6f75_idx"
            ),
        ),
    ]



# Generated by Django 3.2.16 on 2023-01-17 15:01

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0098_pulp_labels'),
        ('ansible', '0047_ansible_namespace'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionVersionMark',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_collectionversionmark', serialize=False, to='core.content')),
                ('value', models.SlugField()),
                ('marked_collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='marks', to='ansible.collectionversion')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('value', 'marked_collection')},
            },
            bases=('core.content',),
        ),
    ]



"""Tests collection sync functionality that is common to both Galaxy V2 and V3."""

import datetime

import pytest

from pulp_ansible.tests.functional.utils import randstr


@pytest.mark.parallel
def test_sync_supports_mirror_option_true(
    ansible_bindings, ansible_collection_remote_factory, ansible_sync_factory
):
    """Sync multiple remotes into the same repo with mirror as `True`."""
    remote_a = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
    )

    remote_b = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote_a.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}1/"
    repository = ansible_sync_factory(repository, remote=remote_b.pulp_href, mirror=True)
    assert repository.latest_version_href == f"{repository.versions_href}2/"

    # Assert more CollectionVersion are present in the first sync than the second
    if repository.retain_repo_versions and repository.retain_repo_versions > 1:
        content_version_one = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content_version_one.count >= 3
    content_version_two = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/2/"
    )
    assert content_version_two.count == 1


@pytest.mark.parallel
def test_sync_supports_mirror_option_false(
    ansible_bindings, ansible_collection_remote_factory, ansible_sync_factory
):
    """Sync multiple remotes into the same repo with mirror as `False`."""
    remote_a = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
    )

    remote_b = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote_a.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}1/"
    repository = ansible_sync_factory(repository, remote=remote_b.pulp_href, mirror=False)
    assert repository.latest_version_href == f"{repository.versions_href}2/"

    # Assert more CollectionVersion are present in the second sync than the first
    if repository.retain_repo_versions and repository.retain_repo_versions > 1:
        content_version_one = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content_version_one.count >= 3
    content_version_two = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/2/"
    )
    assert content_version_two.count == 4


@pytest.mark.parallel
def test_sync_mirror_defaults_to_false(
    ansible_bindings, ansible_collection_remote_factory, ansible_sync_factory
):
    """Sync multiple remotes into the same repo to ensure mirror defaults to `False`."""
    remote_a = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
    )

    remote_b = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote_a.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}1/"
    repository = ansible_sync_factory(repository, remote=remote_b.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}2/"

    # Assert more CollectionVersion are present in the second sync than the first
    if repository.retain_repo_versions and repository.retain_repo_versions > 1:
        content_version_one = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content_version_one.count >= 3
    content_version_two = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/2/"
    )
    assert content_version_two.count == 4


@pytest.mark.parallel
@pytest.mark.parametrize(
    "collection,old_galaxy,sync_dependencies,min_count",
    [
        pytest.param("ibm.ibm_zos_core", False, False, 14, id="with_long_tag"),
        pytest.param("rshad.collection_demo", False, False, 6, id="with_dot_slash_in_manifest"),
        pytest.param("brightcomputing.bcm", True, False, 5, id="with_strange_version_numbers"),
        pytest.param(
            "pulp.pulp_installer",
            False,
            True,
            5,
            id="simple_dependency",
            marks=[pytest.mark.timeout(1800)],  # TODO This test takes too much time!
        ),
    ],
)
def test_sync_collection_with_specialities(
    ansible_bindings,
    collection,
    old_galaxy,
    sync_dependencies,
    min_count,
    ansible_collection_remote_factory,
    ansible_sync_factory,
):
    """Sync a collection that is known to be special."""
    remote = ansible_collection_remote_factory(
        url=(
            "https://old-galaxy.ansible.com/api/"
            if old_galaxy
            else "https://galaxy.ansible.com/api/"
        ),
        requirements_file=f"collections:\n  - {collection}",
        sync_dependencies=sync_dependencies,
    )

    repository = ansible_sync_factory(remote=remote.pulp_href)

    content = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/1/"
    )
    assert content.count >= min_count


class TestFullDependenciesSync:
    @pytest.fixture(scope="class")
    def distribution_with_dependencies(
        self,
        ansible_repo_factory,
        ansible_distribution_factory,
        build_and_upload_collection,
    ):
        """
        Collections with dependencies in a repository.
        """

        # Dependency Graph:
        #     E     A  I
        #     |\    |  |\
        #     F G   B  J K
        #     |  \  |   \|
        #  H(1-3) \ C    L
        #          \|
        #           D

        namespace = randstr()
        TEST_COLLECTION_CONFIGS = [
            {"namespace": namespace, "name": "a", "dependencies": {f"{namespace}.b": "*"}},
            {"namespace": namespace, "name": "b", "dependencies": {f"{namespace}.c": "*"}},
            {"namespace": namespace, "name": "c", "dependencies": {f"{namespace}.d": "*"}},
            {"namespace": namespace, "name": "d"},
            {
                "namespace": namespace,
                "name": "e",
                "dependencies": {f"{namespace}.f": "*", f"{namespace}.g": "*"},
            },
            {"namespace": namespace, "name": "f", "dependencies": {f"{namespace}.h": "<=3.0.0"}},
            {"namespace": namespace, "name": "g", "dependencies": {f"{namespace}.d": "*"}},
            {"namespace": namespace, "name": "h", "version": "1.0.0"},
            {"namespace": namespace, "name": "h", "version": "2.0.0"},
            {"namespace": namespace, "name": "h", "version": "3.0.0"},
            {
                "namespace": namespace,
                "name": "i",
                "dependencies": {f"{namespace}.j": "*", f"{namespace}.k": "*"},
            },
            {"namespace": namespace, "name": "j", "dependencies": {f"{namespace}.l": "*"}},
            {"namespace": namespace, "name": "k", "dependencies": {f"{namespace}.l": "*"}},
            {"namespace": namespace, "name": "l"},
        ]
        repository = ansible_repo_factory()
        distribution = ansible_distribution_factory(
            repository=repository, pulp_labels={"namespace": namespace}
        )
        for config in TEST_COLLECTION_CONFIGS:
            build_and_upload_collection(repository, config={"namespace": namespace, **config})
        return distribution

    @pytest.mark.parametrize(
        "collection_name,expected_count",
        [
            pytest.param("d", 1, id="no_dependency"),
            pytest.param("c", 2, id="simple_one_level"),
            pytest.param("a", 4, id="simple_multi_level"),
            pytest.param("f", 4, id="complex_one_level"),
            pytest.param("e", 7, id="complex_multi_level"),
            pytest.param("i", 4, id="diamond_shaped"),
        ],
    )
    # Running this test in parallel leads to weird failures with upload.
    # @pytest.mark.parallel
    def test_dependency_sync(
        self,
        collection_name,
        expected_count,
        ansible_bindings,
        distribution_with_dependencies,
        ansible_collection_remote_factory,
        ansible_sync_factory,
    ):
        namespace = distribution_with_dependencies.pulp_labels["namespace"]
        remote = ansible_collection_remote_factory(
            url=distribution_with_dependencies.client_url,
            requirements_file=f"collections:\n  - {namespace}.{collection_name}",
            include_pulp_auth=True,
        )
        repository = ansible_sync_factory(remote=remote.pulp_href)

        content = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content.count == expected_count


@pytest.mark.skip("Skipped until fixture metadata has a published date")
@pytest.mark.parallel
def test_optimized_sync(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    monitor_task,
):
    # TODO this test is incomplete and may not work
    remote1 = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    remote2 = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    repository = ansible_repo_factory(remote=remote1.pulp_href)
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {}).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time is not None
    monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {"optimize": True}).task
    )
    # TODO CHECK IF SYNC WAS OPTIMIZED
    monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.sync(
            repository.pulp_href, {"remote": remote2.pulp_href, "optimize": True}
        ).task
    )
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time is None
    # TODO CHECK IF SYNC WAS NOT OPTIMIZED


@pytest.mark.parallel
def test_semver_sync(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    monitor_task,
):
    remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - anil_cm.terraform_provider",
        sync_dependencies=False,
        signed_only=False,
    )
    repository = ansible_repo_factory(remote=remote.pulp_href)
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {}).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    content = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=repository.latest_version_href
    )
    versions = {item.version for item in content.results}
    # If this fails check that it is still upstream!
    # TODO create a better local fixture to sync from.
    assert "0.0.1-rerelease+meta" in versions
    assert "0.0.0-rerelease+meta" in versions


@pytest.mark.parallel
def test_last_synced_metadata_time(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    monitor_task,
):
    remote1 = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    repository = ansible_repo_factory(remote=remote1.pulp_href)
    monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.partial_update(
            repository.pulp_href, {"last_synced_metadata_time": "2000-01-01 00:00"}
        ).task
    )
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time == datetime.datetime(
        2000, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
    )
    monitor_task(
        ansible_bindings.RemotesCollectionApi.partial_update(
            remote1.pulp_href, {"signed_only": True}
        ).task
    )
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time is None


@pytest.mark.parallel
def test_sync_progress_report(
    ansible_bindings,
    ansible_repo,
    ansible_collection_remote_factory,
    monitor_task,
):
    """Checks that the progress report messages are correct."""
    remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    body = {"remote": remote.pulp_href}
    task = monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.sync(ansible_repo.pulp_href, body).task
    )

    progress_reports = task.progress_reports
    assert len(progress_reports) == 5
    messages = {pr.message for pr in progress_reports}

    assert messages == {
        "Parsing CollectionVersion Metadata",
        "Parsing Namespace Metadata",
        "Downloading Artifacts",
        "Downloading Docs Blob",
        "Associating Content",
    }

    for pr in progress_reports:
        if pr.message == "Parsing CollectionVersion Metadata":
            assert pr.total == pr.done
        if pr.message == "Parsing Namespace Metadata":
            assert pr.total == pr.done == 1



from datetime import datetime
from gettext import gettext as _
import semantic_version
import base64
import re

from django.db import DatabaseError, IntegrityError
from django.db.models import F, OuterRef, Exists, Subquery, Prefetch
from django.http import StreamingHttpResponse, HttpResponseNotFound
from django.shortcuts import get_object_or_404, redirect
from django.utils.dateparse import parse_datetime
from django_filters import filters
from django.views.generic.base import RedirectView
from django.conf import settings
from django.core.cache import cache
from django.db.utils import InternalError as DatabaseInternalError

from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from jinja2 import Template
from rest_framework import mixins
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework.reverse import reverse, reverse_lazy
from rest_framework import serializers
from rest_framework import status as http_status
from rest_framework import viewsets, views
from rest_framework.exceptions import NotFound
from rest_framework import status

from pulpcore.plugin.models import (
    Artifact,
    Content,
    ContentArtifact,
    Distribution,
)
from pulpcore.plugin.serializers import AsyncOperationResponseSerializer
from pulpcore.plugin.viewsets import (
    BaseFilterSet,
    OperationPostponedResponse,
    SingleArtifactContentUploadViewSet,
    NAME_FILTER_OPTIONS,
)
from pulpcore.plugin.tasking import add_and_remove, dispatch, general_create
from pulpcore.plugin.util import get_domain, get_domain_pk, get_url

from pulp_ansible.app.galaxy.v3.exceptions import ExceptionHandlerMixin
from pulp_ansible.app.galaxy.v3.serializers import (
    CollectionSerializer,
    CollectionVersionSerializer,
    CollectionVersionDocsSerializer,
    CollectionVersionListSerializer,
    RepoMetadataSerializer,
    UnpaginatedCollectionVersionSerializer,
    ClientConfigurationSerializer,
)
from pulp_ansible.app.models import (
    AnsibleCollectionDeprecated,
    AnsibleDistribution,
    AnsibleNamespaceMetadata,
    Collection,
    CollectionDownloadCount,
    CollectionVersion,
    CollectionVersionMark,
    CollectionVersionSignature,
    CollectionImport,
    DownloadLog,
)
from pulp_ansible.app.serializers import (
    AnsibleNamespaceMetadataSerializer,
    CollectionImportDetailSerializer,
    CollectionOneShotSerializer,
    CollectionVersionUploadSerializer,
)

from pulp_ansible.app.galaxy.mixins import UploadGalaxyCollectionMixin, GalaxyAuthMixin
from pulp_ansible.app.galaxy.v3.pagination import LimitOffsetPagination
from pulp_ansible.app.viewsets import (
    CollectionVersionFilter,
)

from pulp_ansible.app.tasks.deletion import delete_collection_version, delete_collection

from pulp_ansible.app.utils import filter_content_for_repo_version


DOMAIN_ENABLED = settings.DOMAIN_ENABLED

_CAN_VIEW_REPO_CONTENT = {
    "action": ["list", "retrieve", "download"],
    "principal": "authenticated",
    "effect": "allow",
    "condition": "v3_can_view_repo_content",
}

_PERMISSIVE_ACCESS_POLICY = {
    "statements": [
        _CAN_VIEW_REPO_CONTENT,
    ],
    "creation_hooks": [],
}


def encode_cache_key(key):
    """
    Base64 encode the cache key to avoid any unpleasantness with unsupported
    characters.
    """
    b64 = base64.b64encode(key.encode())
    return b64.decode()


class AnsibleDistributionMixin:
    """
    A mixin for ViewSets that use AnsibleDistribution.
    """

    _repo_version = None
    _repo = None

    @property
    def _repository_version(self):
        """Returns repository version."""
        if self._repo_version:
            return self._repo_version

        path = self.kwargs["distro_base_path"]

        context = getattr(self, "pulp_context", None)
        if context and context.get(path, None):
            return self.pulp_context[path]

        # using Distribution instead of Ansible distribution allows us to save a join
        distro = get_object_or_404(
            Distribution.objects.only(
                "base_path", "repository_version", "repository"
            ).select_related("repository__ansible_ansiblerepository"),
            base_path=path,
            pulp_domain=get_domain_pk(),
        )
        if distro.repository_version_id:
            self.pulp_context = {path: distro.repository_version}
            self._repo_version = distro.repository_version
            self._repo = self._repo_version.repository.cast()
            return distro.repository_version

        repo_version = distro.repository.latest_version() if distro.repository else None
        self.pulp_context = {path: repo_version}
        self._repo = repo_version.repository.cast() if repo_version else None
        self._repo_version = repo_version
        return repo_version

    @property
    def _repository(self):
        if self._repo:
            return self._repo
        self._repo_version
        return self._repo

    @property
    def _distro_content(self):
        """Returns distribution content."""
        repo_version = self._repository_version
        if repo_version is None:
            return Content.objects.none()

        return repo_version.content

    def get_serializer_context(self):
        """Inserts distribution path to a serializer context."""
        context = super().get_serializer_context()
        if "path" in self.kwargs:
            context["path"] = self.kwargs["path"]

        context["distro_base_path"] = self.kwargs["distro_base_path"]
        return context


class CollectionVersionRetrieveMixin:
    """
    A mixin for ViewSets that get instance of CollectionVersion.
    """

    def get_queryset(self):
        """
        Returns a CollectionVersions queryset for specified distribution.
        """
        if getattr(self, "swagger_fake_view", False):
            # drf_spectacular get filter from get_queryset().model
            # and it fails when "path" is not on self.kwargs
            return CollectionVersion.objects.none()

        repo_version = self._repository_version

        qs = (
            filter_content_for_repo_version(CollectionVersion.objects, repo_version)
            .select_related("collection")
            .prefetch_related(
                Prefetch(
                    "marks",
                    queryset=filter_content_for_repo_version(
                        CollectionVersionMark.objects, repo_version
                    ),
                )
            )
            .filter(namespace=self.kwargs["namespace"], name=self.kwargs["name"])
        )

        return qs

    def get_object(self):
        """
        This modifies the qs for get requests to add some extra lookups for the detail view
        """
        repo_version = self._repository_version
        qs = self.get_queryset()
        qs = (
            qs.prefetch_related(
                Prefetch(
                    "signatures",
                    queryset=filter_content_for_repo_version(
                        CollectionVersionSignature.objects, repo_version
                    ),
                )
            )
            .prefetch_related(
                Prefetch(
                    "contentartifact_set",
                    queryset=ContentArtifact.objects.select_related("artifact"),
                    to_attr="artifacts",
                )
            )
            .annotate(
                namespace_sha256=Subquery(
                    filter_content_for_repo_version(AnsibleNamespaceMetadata.objects, repo_version)
                    .filter(name=OuterRef("namespace"))
                    .values("metadata_sha256"),
                )
            )
            .select_related("collection")
        )

        queryset = self.filter_queryset(qs)

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            "Expected view %s to be called with a URL keyword argument "
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            "attribute on the view correctly." % (self.__class__.__name__, lookup_url_kwarg)
        )

        filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
        obj = get_object_or_404(queryset, **filter_kwargs)

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj


class CollectionFilter(BaseFilterSet):
    """
    FilterSet for Ansible Collections.
    """

    namespace = filters.CharFilter(field_name="namespace")
    name = filters.CharFilter(field_name="name")
    deprecated = filters.BooleanFilter(field_name="deprecated")

    class Meta:
        model = Collection
        fields = ["namespace", "name", "deprecated"]


class CollectionViewSet(
    GalaxyAuthMixin,
    ExceptionHandlerMixin,
    AnsibleDistributionMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
):
    """
    ViewSet for Collections.
    """

    serializer_class = CollectionSerializer
    filterset_class = CollectionFilter
    pagination_class = LimitOffsetPagination

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            _CAN_VIEW_REPO_CONTENT,
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:ansible.delete_collection",
            },
            {
                "action": ["update", "partial_update"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "v3_can_modify_repo_content",
            },
        ],
    }

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collections"

    def get_queryset(self):
        """
        Returns a Collections queryset for specified distribution.
        """
        if getattr(self, "swagger_fake_view", False):
            # drf_spectacular get filter from get_queryset().model
            # and it fails when "path" is not on self.kwargs
            return Collection.objects.none()
        repo_version = self._repository_version

        deprecated_qs = filter_content_for_repo_version(
            AnsibleCollectionDeprecated.objects, repo_version
        ).filter(namespace=OuterRef("namespace"), name=OuterRef("name"))

        latest_cv_version_qs = (
            filter_content_for_repo_version(CollectionVersion.objects, repo_version)
            .filter(collection=OuterRef("pk"))
            .order_by(
                "-version_major",
                "-version_minor",
                "-version_patch",
                "-version_prerelease",
                "-pulp_created",
            )
            .only("version")
        )

        download_count_qs = CollectionDownloadCount.objects.filter(
            name=OuterRef("name"), namespace=OuterRef("namespace"), pulp_domain=get_domain()
        )

        qs = (
            Collection.objects.annotate(
                highest_version=Subquery(latest_cv_version_qs.values("version")[:1]),
                latest_version_modified=Subquery(
                    latest_cv_version_qs.values("pulp_last_updated")[:1]
                ),
            )
            .annotate(
                deprecated=Exists(deprecated_qs),
                download_count=Subquery(download_count_qs.values("download_count")[:1]),
            )
            .filter(highest_version__isnull=False)
        )

        return qs

    def get_object(self):
        """
        Returns a Collection object.
        """
        queryset = self.filter_queryset(self.get_queryset())
        return get_object_or_404(
            queryset, namespace=self.kwargs["namespace"], name=self.kwargs["name"]
        )

    @extend_schema(
        description="Trigger an asynchronous update task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def update(self, request, *args, **kwargs):
        """
        Update a Collection object.
        """
        repo_version = self._repository_version
        collection = self.get_object()
        serializer = self.get_serializer(collection, data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)
        deprecated_value = request.data.get("deprecated")
        add_content_units = []
        remove_content_units = []

        deprecation, created = AnsibleCollectionDeprecated.objects.get_or_create(
            namespace=collection.namespace, name=collection.name
        )
        if not created:
            try:
                deprecation.touch()
            except DatabaseError:
                # deprecation has since been removed. try to create it
                deprecation = AnsibleCollectionDeprecated.objects.create(
                    namespace=collection.namespace, name=collection.name
                )

        if deprecated_value:
            add_content_units.append(deprecation.pk)
        else:
            remove_content_units.append(deprecation.pk)

        task = dispatch(
            add_and_remove,
            exclusive_resources=[repo_version.repository],
            kwargs={
                "repository_pk": repo_version.repository.pk,
                "base_version_pk": repo_version.pk,
                "add_content_units": add_content_units,
                "remove_content_units": remove_content_units,
            },
        )
        return OperationPostponedResponse(task, request)

    @extend_schema(
        description="Trigger an asynchronous delete task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def destroy(self, request: Request, *args, **kwargs) -> Response:
        """
        Allow a Collection to be deleted.

        1. Perform Dependency Check to verify that each CollectionVersion
           inside Collection can be deleted
        2. If the Collection can’t be deleted, return the reason why
        3. If it can, dispatch task to delete each CollectionVersion
           and the Collection
        """
        collection = self.get_object()

        # dependency check
        dependents = get_collection_dependents(collection)
        if dependents:
            return Response(
                {
                    "detail": _(
                        "Collection {namespace}.{name} could not be deleted "
                        "because there are other collections that require it."
                    ).format(
                        namespace=collection.namespace,
                        name=collection.name,
                    ),
                    "dependent_collection_versions": [
                        f"{dep.namespace}.{dep.name} {dep.version}" for dep in dependents
                    ],
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        repositories = set()
        for version in collection.versions.all():
            for repo in version.repositories.all():
                repositories.add(repo)

        async_result = dispatch(
            delete_collection,
            exclusive_resources=list(repositories),
            kwargs={"collection_pk": collection.pk},
        )

        return OperationPostponedResponse(async_result, request)


def get_collection_dependents(parent):
    """Given a parent collection, return a list of collection versions that depend on it."""
    key = f"{parent.namespace}.{parent.name}"
    return list(
        CollectionVersion.objects.exclude(collection=parent).filter(
            dependencies__has_key=key, pulp_domain=get_domain()
        )
    )


def get_unique_dependents(parent):
    """Given a parent collection version, return a list of collection versions that depend on it."""
    key = f"{parent.namespace}.{parent.name}"

    this_version = semantic_version.Version(parent.version)

    # Other versions contain a set of all versions of this collection aside from the version
    # that is being deleted.
    other_versions = []
    for v in parent.collection.versions.exclude(version=parent.version):
        other_versions.append(semantic_version.Version(v.version))

    dependents = []
    for child in CollectionVersion.objects.filter(
        dependencies__has_key=key, pulp_domain=get_domain()
    ):
        spec = semantic_version.SimpleSpec(child.dependencies[key])

        # If this collection matches the parent collections version and there are no other
        # collection versions that can satisfy the requirement, add it to the list of dependants.
        if spec.match(this_version) and not spec.select(other_versions):
            dependents.append(child)

    return dependents


class UnpaginatedCollectionViewSet(CollectionViewSet):
    """Unpaginated ViewSet for Collections."""

    pagination_class = None

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collections/all"


class CollectionUploadViewSet(
    GalaxyAuthMixin,
    ExceptionHandlerMixin,
    UploadGalaxyCollectionMixin,
    SingleArtifactContentUploadViewSet,
    AnsibleDistributionMixin,
):
    """
    ViewSet for Collection Uploads.
    """

    queryset = None
    endpoint_pieces = None
    serializer_class = CollectionVersionUploadSerializer
    pulp_tag_name = "Pulp_Ansible: Artifacts Collections V3"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            _CAN_VIEW_REPO_CONTENT,
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_perms:ansible.add_collection",
                    "v3_can_view_repo_content",
                ],
            },
        ],
    }

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collections/upload"

    def _dispatch_upload_collection_task(self, args=None, kwargs=None, repository=None):
        """
        Dispatch an Upload Collection creation task.
        """
        locks = []
        if repository:
            locks.append(repository)

        return dispatch(general_create, exclusive_resources=locks, args=args, kwargs=kwargs)

    @extend_schema(
        description="Create an artifact and trigger an asynchronous task to create "
        "Collection content from it.",
        summary="Upload a collection",
        request=CollectionOneShotSerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request, distro_base_path):
        """
        Dispatch a Collection creation task.
        """
        distro = get_object_or_404(AnsibleDistribution, base_path=distro_base_path)
        repo = distro.repository
        if repo is None:
            if distro.repository_version is None:
                raise serializers.ValidationError(
                    _("Distribution must have either repository or repository_version set")
                )
            repo = distro.repository_version.repository
        # Check that invalid fields were not specified
        serializer = CollectionOneShotSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # Check that namespace, name and version can be extracted
        request.data["repository"] = get_url(repo)
        serializer = CollectionVersionUploadSerializer(
            data=request.data, context=self.get_serializer_context()
        )
        serializer.is_valid(raise_exception=True)
        # Convert file to an artifact
        task_payload = self.init_content_data(serializer, request)
        # Dispatch create task
        task = self._dispatch_upload_collection_task(
            repository=serializer.validated_data["repository"],
            args=(CollectionVersion._meta.app_label, serializer.__class__.__name__),
            kwargs={
                "data": task_payload,
                "context": self.get_deferred_context(request),
            },
        )
        # Create CollectionImport and response
        CollectionImport.objects.create(task_id=task.pk)
        if DOMAIN_ENABLED:
            kwargs = {"pk": task.pk, "pulp_domain": get_domain().name}
        else:
            kwargs = {"pk": task.pk}
        data = {
            "task": reverse(
                settings.ANSIBLE_URL_NAMESPACE + "collection-imports-detail",
                kwargs=kwargs,
                request=None,
            )
        }
        return Response(data, status=http_status.HTTP_202_ACCEPTED)


class LegacyCollectionUploadViewSet(CollectionUploadViewSet):
    """
    Collection upload viewset with deprecated markers for the openAPI spec.
    """

    @extend_schema(
        description="Create an artifact and trigger an asynchronous task to create "
        "Collection content from it.",
        summary="Upload a collection",
        request=CollectionOneShotSerializer,
        responses={202: AsyncOperationResponseSerializer},
        deprecated=True,
    )
    def create(self, request, path):
        """Create collection."""
        return super().create(request, distro_base_path=path)


class CollectionArtifactDownloadView(GalaxyAuthMixin, views.APIView, AnsibleDistributionMixin):
    """Collection download endpoint."""

    action = "download"

    DEFAULT_ACCESS_POLICY = _PERMISSIVE_ACCESS_POLICY

    @staticmethod
    def log_download(request, namespace, name, version, distro_base_path):
        """Log the download of the collection version."""

        def _get_org_id(request):
            if not isinstance(request.auth, dict):
                return None

            x_rh_identity = request.auth.get("rh_identity")
            if not x_rh_identity:
                return None

            identity = x_rh_identity["identity"]

            if (not identity) or (not identity.get("internal")):
                return None

            return identity["internal"]["org_id"]

        # Get user IP
        x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
        ip = x_forwarded_for.split(",")[0] if x_forwarded_for else request.META.get("REMOTE_ADDR")

        user_agent = request.headers.get("user-agent", "unknown")

        distribution = get_object_or_404(AnsibleDistribution, base_path=distro_base_path)
        repository_version = (
            distribution.repository_version or distribution.repository.latest_version()
        )

        collection_version = get_object_or_404(
            CollectionVersion.objects.filter(pk__in=repository_version.content),
            namespace=namespace,
            name=name,
            version=version,
        )

        log_params = dict(
            content_unit=collection_version,
            ip=ip,
            extra_data={"org_id": _get_org_id(request)},
            user_agent=user_agent,
            repository=repository_version.repository,
            repository_version=repository_version,
        )
        if request.user.is_authenticated:
            log_params["user"] = request.user

        try:
            DownloadLog.objects.create(**log_params)
        except DatabaseInternalError as e:
            # handle a read-only replica scenario
            msg = str(e)
            if "read-only" in msg:
                return
            raise

    def count_download(namespace, name):
        try:
            collection, created = CollectionDownloadCount.objects.get_or_create(
                namespace=namespace, name=name, defaults={"download_count": 1}
            )
            if not created:
                collection.download_count = F("download_count") + 1
                collection.save()
        except DatabaseInternalError as e:
            # handle a read-only replica scenario
            msg = str(e)
            if "read-only" in msg:
                return
            raise

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collections/download"

    def get(self, request, *args, **kwargs):
        """Download collection."""
        distro_base_path = self.kwargs["distro_base_path"]
        distribution = AnsibleDistribution.objects.get(
            base_path=distro_base_path, pulp_domain=get_domain()
        )

        url = "{host}/{prefix}{domain}/{distro_base_path}/{filename}".format(
            host=settings.CONTENT_ORIGIN.strip("/"),
            prefix=settings.CONTENT_PATH_PREFIX.strip("/"),
            domain="/" + get_domain().name if DOMAIN_ENABLED else "",
            distro_base_path=distro_base_path,
            filename=self.kwargs["filename"],
        )

        match = re.fullmatch(
            r"(?P<namespace>\w+)-(?P<name>\w+)-(?P<version>[\w.-]+)\.tar\.gz",
            self.kwargs["filename"],
        )
        if not match:
            raise NotFound()
        namespace = match.group("namespace")
        name = match.group("name")
        version = match.group("version")

        if settings.ANSIBLE_COLLECT_DOWNLOAD_LOG:
            CollectionArtifactDownloadView.log_download(
                request, namespace, name, version, distro_base_path
            )

        if settings.ANSIBLE_COLLECT_DOWNLOAD_COUNT:
            CollectionArtifactDownloadView.count_download(namespace, name)

        if (
            distribution.content_guard
            and distribution.content_guard.pulp_type == "core.content_redirect"
        ):
            guard = distribution.content_guard.cast()
            url = guard.preauthenticate_url(url)

        return redirect(url)


@extend_schema_view(
    create=extend_schema(responses={202: AsyncOperationResponseSerializer}),
    partial_update=extend_schema(responses={202: AsyncOperationResponseSerializer}),
    delete=extend_schema(responses={202: AsyncOperationResponseSerializer}),
)
class AnsibleNamespaceViewSet(
    GalaxyAuthMixin, ExceptionHandlerMixin, AnsibleDistributionMixin, viewsets.ModelViewSet
):
    serializer_class = AnsibleNamespaceMetadataSerializer
    lookup_field = "name"
    filterset_fields = {
        "name": NAME_FILTER_OPTIONS,
        "company": NAME_FILTER_OPTIONS,
        "metadata_sha256": ["exact", "in"],
    }

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            _CAN_VIEW_REPO_CONTENT,
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "conditions": "v3_can_modify_repo_content",
            },
            {
                "action": "delete",
                "principal": "authenticated",
                "effect": "allow",
                "conditions": "v3_can_modify_repo_content",
            },
            {
                "action": "partial_update",
                "principal": "authenticated",
                "effect": "allow",
                "conditions": "v3_can_modify_repo_content",
            },
        ],
    }

    def get_queryset(self):
        if getattr(self, "swagger_fake_view", False):
            # drf_spectacular get filter from get_queryset().model
            # and it fails when "path" is not on self.kwargs
            return AnsibleNamespaceMetadata.objects.none()
        return filter_content_for_repo_version(
            AnsibleNamespaceMetadata.objects, self._repository_version
        )

    def create(self, request, *args, **kwargs):
        return self._create(request, data=request.data)

    def _create(self, request, data, context=None):
        """Dispatch task to create and add Namespace to repository."""
        repo = self._repository_version.repository
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)

        context = context or {}
        context["repository"] = repo.pk
        # If avatar was uploaded, init into artifact
        if avatar := serializer.validated_data.pop("avatar", None):
            artifact = Artifact.init_and_validate(avatar)
            try:
                artifact.save()
            except IntegrityError:
                # if artifact already exists, let's use it
                try:
                    artifact = Artifact.objects.get(
                        sha256=artifact.sha256, pulp_domain=get_domain()
                    )
                    artifact.touch()
                except (Artifact.DoesNotExist, DatabaseError):
                    # the artifact has since been removed from when we first attempted to save it
                    artifact.save()
            context["artifact"] = artifact.pk

        # Dispatch general_create task
        app_label = AnsibleNamespaceMetadata._meta.app_label
        task = dispatch(
            general_create,
            args=(app_label, serializer.__class__.__name__),
            exclusive_resources=[repo],
            kwargs={
                "data": serializer.validated_data,
                "context": context,
            },
        )
        return OperationPostponedResponse(task, request)

    def update(self, request, *args, **kwargs):
        """Dispatch task to update Namespace in repository."""
        partial = kwargs.pop("partial", False)
        namespace = self.get_object()
        serializer = self.get_serializer(namespace, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)

        for name, field in serializer.fields.items():
            if not field.read_only and not field.write_only:
                serializer.validated_data.setdefault(name, serializer.data[name])
        context = {}
        if "avatar" not in request.data and namespace.avatar_sha256:
            context["artifact"] = Artifact.objects.get(
                sha256=namespace.avatar_sha256, pulp_domain=get_domain()
            ).pk
        return self._create(request, data=serializer.validated_data, context=context)

    def delete(self, request, *args, **kwargs):
        """Try to remove the Namespace if no Collections under Namespace are present."""
        namespace = self.get_object()

        if self._distro_content.filter(
            ansible_collectionversion__namespace=namespace.name
        ).exists():
            raise serializers.ValidationError(
                detail=_(
                    "Namespace {name} cannot be deleted because "
                    "there are still collections associated with it."
                ).format(name=namespace.name)
            )

        repository = self._repository_version.repository
        async_result = dispatch(
            add_and_remove,
            args=(repository.pk, [], [namespace.pk]),
            exclusive_resources=[repository],
        )
        return OperationPostponedResponse(async_result, request)


class CollectionVersionViewSet(
    GalaxyAuthMixin,
    CollectionVersionRetrieveMixin,
    ExceptionHandlerMixin,
    AnsibleDistributionMixin,
    viewsets.GenericViewSet,
):
    """
    ViewSet for CollectionVersions.
    """

    serializer_class = CollectionVersionSerializer
    list_serializer_class = CollectionVersionListSerializer
    filterset_class = CollectionVersionFilter
    pagination_class = LimitOffsetPagination

    lookup_field = "version"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            _CAN_VIEW_REPO_CONTENT,
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:ansible.delete_collection",
            },
        ],
    }

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collection-versions"

    def get_list_serializer(self, *args, **kwargs):
        """
        Return the list serializer instance.
        """
        kwargs.setdefault("context", self.get_serializer_context)
        return self.list_serializer_class(*args, **kwargs)

    @extend_schema(
        responses={200: list_serializer_class},
    )
    def list(self, request, *args, **kwargs):
        """
        Returns paginated CollectionVersions list.
        """

        # there are too many potential permutations of query params to be able
        # to cache all of them, so this will only cache queries that stick to
        # limit/offset params, which the ansible-galaxy client uses
        cache_request = True
        for k in request.query_params.keys():
            if k not in ("offset", "limit"):
                cache_request = False
                break

        cache_key = encode_cache_key(
            "-".join(
                [
                    settings.ANSIBLE_API_HOSTNAME,
                    str(self._repository_version.pk),
                    kwargs["namespace"],
                    kwargs["name"],
                    str(request.query_params.get("offset", "0")),
                    str(request.query_params.get("limit", "0")),
                ]
            )
        )

        if cache_key in cache and cache_request:
            return Response(cache.get(cache_key))

        queryset = self.get_queryset()

        # prevent OOMKILL ...
        queryset = queryset.only(
            "pk",
            "content_ptr_id",
            "marks",
            "namespace",
            "name",
            "version",
            "pulp_created",
            "pulp_last_updated",
            "requires_ansible",
            "collection",
        )

        queryset = self.filter_queryset(queryset)

        context = self.get_serializer_context()
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_list_serializer(page, many=True, context=context)
            data = self.paginator.get_paginated_data(serializer.data)
            if cache_request:
                cache.set(cache_key, data)

            return Response(data)

        serializer = self.get_list_serializer(queryset, many=True, context=context)
        return Response(serializer.data)

    def retrieve(self, request, *args, **kwargs):
        """
        Returns a CollectionVersion object.
        """
        repo_version = self._repository_version
        # the contents of a repo version can be cached without worry because repo
        # versions are immutable
        cache_key = encode_cache_key(
            f"{settings.ANSIBLE_API_HOSTNAME}-version-details-{repo_version.pk}"
            f"-{kwargs['namespace']}-{kwargs['name']}-{kwargs['version']}"
        )

        if data := cache.get(cache_key):
            return Response(data)

        instance = self.get_object()
        context = self.get_serializer_context()
        serializer = self.get_serializer_class()(instance, context=context)
        cache.set(cache_key, serializer.data)

        return Response(serializer.data)

    @extend_schema(
        description="Trigger an asynchronous delete task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def destroy(self, request: Request, *args, **kwargs) -> Response:
        """
        Allow a CollectionVersion to be deleted.

        1. Perform Dependency Check to verify that the collection version can be deleted
        2. If the collection version can’t be deleted, return the reason why
        3. If it can, dispatch task to delete CollectionVersion and clean up repository.
           If the version being deleted is the last collection version in the collection,
           remove the collection object as well.
        """
        collection_version = self.get_object()

        # dependency check
        dependents = get_unique_dependents(collection_version)
        if dependents:
            return Response(
                {
                    "detail": _(
                        "Collection version {namespace}.{name} {version} could not be "
                        "deleted because there are other collections that require it."
                    ).format(
                        namespace=collection_version.namespace,
                        name=collection_version.collection.name,
                        version=collection_version.version,
                    ),
                    "dependent_collection_versions": [
                        f"{dep.namespace}.{dep.name} {dep.version}" for dep in dependents
                    ],
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        async_result = dispatch(
            delete_collection_version,
            exclusive_resources=collection_version.repositories.all(),
            kwargs={"collection_version_pk": collection_version.pk},
        )

        return OperationPostponedResponse(async_result, request)


class UnpaginatedCollectionVersionViewSet(CollectionVersionViewSet):
    """Unpaginated ViewSet for CollectionVersions."""

    list_serializer_class = UnpaginatedCollectionVersionSerializer
    pagination_class = None

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collection-versions/all"

    def get_queryset(self):
        """
        Returns a CollectionVersions queryset for specified distribution.
        """
        return (
            filter_content_for_repo_version(
                CollectionVersion.objects.select_related(), self._repository_version
            )
            .annotate(
                namespace_sha256=Subquery(
                    filter_content_for_repo_version(
                        AnsibleNamespaceMetadata.objects, self._repository_version
                    )
                    .filter(name=OuterRef("namespace"))
                    .values("metadata_sha256"),
                )
            )
            .prefetch_related(
                Prefetch(
                    "contentartifact_set",
                    queryset=ContentArtifact.objects.select_related("artifact"),
                    to_attr="artifacts",
                )
            )
            .select_related("collection")
        )

    @extend_schema(
        responses={200: list_serializer_class},
    )
    def list(self, request, *args, **kwargs):
        """
        Returns paginated CollectionVersions list.
        """
        queryset = self.get_queryset().iterator(chunk_size=100)

        context = self.get_serializer_context()
        cvs_template_string = (
            "[{% for cv in versions %}"
            "{{ cv|tojson }}{% if not loop.last %},{% endif %}"
            "{% endfor %}]"
        )
        cvs_template = Template(cvs_template_string)
        serialized_map = (self.get_list_serializer(x, context=context).data for x in queryset)
        streamed = (x.encode("utf-8") for x in cvs_template.stream(versions=serialized_map))
        return StreamingHttpResponse(streamed)


class CollectionVersionDocsViewSet(
    GalaxyAuthMixin,
    CollectionVersionRetrieveMixin,
    ExceptionHandlerMixin,
    AnsibleDistributionMixin,
    viewsets.GenericViewSet,
):
    """
    ViewSet for docs_blob of CollectionVersion.
    """

    serializer_class = CollectionVersionDocsSerializer

    lookup_field = "version"

    DEFAULT_ACCESS_POLICY = _PERMISSIVE_ACCESS_POLICY

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collection-versions/docs"

    def retrieve(self, request, *args, **kwargs):
        """
        Returns a CollectionVersion object.
        """
        repo_version = self._repository_version
        # the contents of a repo version can be cached without worry because repo
        # versions are immutable
        cache_key = encode_cache_key(
            f"version-docs-{repo_version.pk}-"
            f"{kwargs['namespace']}-{kwargs['name']}-{kwargs['version']}"
        )

        if data := cache.get(cache_key):
            return Response(data)

        instance = self.get_object()
        context = self.get_serializer_context()
        serializer = self.get_serializer_class()(instance, context=context)
        cache.set(cache_key, serializer.data)

        return Response(serializer.data)


class CollectionImportViewSet(
    GalaxyAuthMixin,
    ExceptionHandlerMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
    AnsibleDistributionMixin,
):
    """
    ViewSet for CollectionImports.
    """

    queryset = CollectionImport.objects.prefetch_related("task").all()
    serializer_class = CollectionImportDetailSerializer

    DEFAULT_ACCESS_POLICY = _PERMISSIVE_ACCESS_POLICY

    since_filter = OpenApiParameter(
        name="since",
        location=OpenApiParameter.QUERY,
        type=str,
        # format=openapi.FORMAT_DATETIME,
        description="Filter messages since a given timestamp",
    )

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/collections/imports"

    @extend_schema(parameters=[since_filter])
    def retrieve(self, request, *args, **kwargs):
        """
        Returns a CollectionImport object.
        """
        instance = self.get_object()

        if "since" in self.request.query_params:
            since = parse_datetime(self.request.query_params["since"])
            messages = []
            for message in instance.messages:
                message_time = datetime.fromtimestamp(message["time"])
                if message_time.replace(tzinfo=since.tzinfo) > since:
                    messages.append(message)
            instance.messages = messages

        context = self.get_serializer_context()
        serializer = CollectionImportDetailSerializer(instance, context=context)

        return Response(serializer.data)


class RepoMetadataViewSet(
    GalaxyAuthMixin,
    ExceptionHandlerMixin,
    AnsibleDistributionMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
):
    """
    ViewSet for Repository Metadata.
    """

    DEFAULT_ACCESS_POLICY = _PERMISSIVE_ACCESS_POLICY

    serializer_class = RepoMetadataSerializer

    def urlpattern(*args, **kwargs):
        """Return url pattern for RBAC."""
        return "pulp_ansible/v3/repo-metadata"

    def get_object(self):
        """
        Returns a RepositoryVersion object.
        """
        return self._repository_version


def redirect_view_generator(actions, url, viewset, distro_view=True, responses={}):
    """
    Generates a redirect view.

    Generates a viewset that
        - Redirects the selected actions to the given URL
        - Mocks the OpenAPI generation for the selected viewset

    Args:
        actions: dictionary of actions to be passed into .as_view()
        url: path name of the viewset that this view should redirect to
        viewset: viewset class of the viewset that this path redirects to
        distro_view: indicates if the viewset that is being redirected to is scoped to
            a distribution base path (such as collections).
        responses: allows the 202 response type to be overridden for specific actions.
            ex: responses={"list": MyListSerializer} will override the list action on
            the viewset to return MyListSerializer in the open api spec

    Returns:
        ViewSet.as_view()
    """
    # Serializers from the parent class are required so that the auto generated
    # clients return the correct data structure when they call the API.
    serializer = getattr(viewset, "serializer_class", None)
    list_serializer = getattr(viewset, "list_serializer_class", serializer)

    # TODO: it would be nice to be able to get a url pattern for the view that
    # this redirects to to add to the description.
    # url_pattern = resolve(settings.ANSIBLE_URL_NAMESPACE + url).route

    description = "Legacy v3 endpoint."

    # TODO: Might be able to load serializer info directly from spectacular
    # schema that gets set on the viewset
    def get_responses(action):
        default = serializer
        if action == "list":
            default = list_serializer
        elif action == "destroy":
            default = AsyncOperationResponseSerializer
        response = responses.get(action, default)
        if response == AsyncOperationResponseSerializer:
            status_code = 202
        elif response is None:
            status_code = 204
        else:
            status_code = 200
        return {status_code: response, 302: None}

    # subclasses viewset to make .as_view work correctly on non viewset views
    # subclasses the redirected viewset to get pagination class and query params
    class GeneratedRedirectView(RedirectView, viewsets.ViewSet, viewset):
        permanent = False

        def urlpattern(*args, **kwargs):
            """Return url pattern for RBAC."""
            return "pulp_ansible/v3/legacy-redirected-viewset"

        def _get(self, *args, **kwargs):
            try:
                return super().get(*args, **kwargs)
            except NotFound:
                return HttpResponseNotFound()

        def get_redirect_url(self, *args, **kwargs):
            if distro_view:
                if "path" not in kwargs:
                    if settings.ANSIBLE_DEFAULT_DISTRIBUTION_PATH is None:
                        raise NotFound()
                    else:
                        path = settings.ANSIBLE_DEFAULT_DISTRIBUTION_PATH
                else:
                    path = self.kwargs["path"]
                    # remove the old path kwarg since we're redirecting to the new api endpoints
                    del kwargs["path"]

                kwargs["distro_base_path"] = path

            if DOMAIN_ENABLED:
                kwargs["pulp_domain"] = get_domain().name

            # don't pass request. redirects work with just the path and this solves the client
            # redirect issues for aiohttp
            url = reverse_lazy(
                settings.ANSIBLE_URL_NAMESPACE + self.url,
                kwargs=kwargs,
                # request=self.request,
            )

            qs = self.request.META.get("QUERY_STRING", "")
            if qs:
                url = "%s?%s" % (url, qs)

            return url

        @extend_schema(
            description=description,
            responses=get_responses("retrieve"),
            deprecated=True,
        )
        def retrieve(self, request, *args, **kwargs):
            return self._get(request, *args, **kwargs)

        @extend_schema(
            description=description,
            responses=get_responses("destroy"),
            deprecated=True,
        )
        def destroy(self, request, *args, **kwargs):
            return self._get(request, *args, **kwargs)

        @extend_schema(
            description=description,
            responses=get_responses("list"),
            deprecated=True,
        )
        def list(self, request, *args, **kwargs):
            return self._get(request, *args, **kwargs)

        @extend_schema(
            description=description,
            responses=get_responses("update"),
            deprecated=True,
        )
        def update(self, request, *args, **kwargs):
            return self._get(request, *args, **kwargs)

    return GeneratedRedirectView.as_view(actions, url=url)


class ClientConfigurationView(GalaxyAuthMixin, views.APIView, AnsibleDistributionMixin):
    """Return configurations for the ansible-galaxy client."""

    DEFAULT_ACCESS_POLICY = _PERMISSIVE_ACCESS_POLICY

    action = "retrieve"

    @extend_schema(responses=ClientConfigurationSerializer)
    def get(self, request, *args, **kwargs):
        """Get the client configs."""

        data = ClientConfigurationSerializer(
            {
                "default_distribution_path": self.kwargs.get(
                    "path", settings.ANSIBLE_DEFAULT_DISTRIBUTION_PATH
                )
            }
        )

        return Response(data.data)



# Generated by Django 3.2.14 on 2022-07-21 22:35
from django.db import migrations


# This migration is supposed to be zero downtime safe.
# It will attempt to populate the sha256 field wherever possible. But it is expected that wherever
# older versions pulp_ansible are still running, CollectionVersions without this digest will be
# created.


def add_sha256_to_current_models(apps, schema_editor):
    """Adds the sha256 to current CollectionVersion models."""
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")
    collection_versions_to_update = []
    collection_versions_on_demand = []

    for collection_version in (
        CollectionVersion.objects.prefetch_related(
            "contentartifact_set", "contentartifact_set__artifact"
        )
        .filter(sha256="")
        .only("pk", "sha256")
        .iterator(chunk_size=2000)
    ):
        content_artifact = collection_version.contentartifact_set.get()
        if content_artifact.artifact:
            collection_version.sha256 = content_artifact.artifact.sha256
            collection_versions_to_update.append(collection_version)
        else:
            collection_versions_on_demand.append(collection_version)
        if len(collection_versions_to_update) >= 1024:
            CollectionVersion.objects.bulk_update(
                collection_versions_to_update,
                [
                    "sha256",
                ],
            )
            collection_versions_to_update.clear()
    # Update remaining collection versions
    if len(collection_versions_to_update) > 0:
        CollectionVersion.objects.bulk_update(
            collection_versions_to_update,
            [
                "sha256",
            ],
        )

    # If there are on-demand collections then the next migration will fail, so error here with
    # helpful message on how to fix. No work will be performed by this migration on a second-run.
    if len(collection_versions_on_demand) > 0:
        raise Exception(
            f"On demand collections found. Please remove or upload/sync their data: "
            f"{[c.pk for c in collection_versions_on_demand]}"
        )


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0056_collectionversion_sha256"),
    ]

    operations = [
        migrations.RunPython(add_sha256_to_current_models, migrations.RunPython.noop, elidable=True)
    ]



import json

from import_export import fields
from import_export.widgets import Widget
from pulpcore.plugin.importexport import BaseContentResource, QueryModelResource
from pulp_ansible.app.models import (
    AnsibleCollectionDeprecated,
    AnsibleNamespace,
    AnsibleNamespaceMetadata,
    CollectionVersionMark,
    Role,
    Collection,
    CollectionVersion,
    CollectionVersionSignature,
)


class RoleContentResource(BaseContentResource):
    """
    Resource for import/export of ansible_role-content entities.
    """

    def set_up_queryset(self):
        """
        :return: Role content specific to a specified repo-version.
        """
        return Role.objects.filter(pk__in=self.repo_version.content)

    class Meta:
        model = Role
        import_id_fields = model.natural_key_fields()


class AnsibleNamespaceResource(QueryModelResource):
    """
    Resource for import/export of ansible_namespace entities
    """

    def set_up_queryset(self):
        """
        :return: Collections specific to a specified repo-version.
        """
        namespace_metadatas = AnsibleNamespaceMetadata.objects.filter(
            pk__in=self.repo_version.content
        )
        return AnsibleNamespace.objects.filter(
            pk__in=namespace_metadatas.values_list("namespace", flat=True)
        )

    class Meta:
        model = AnsibleNamespace
        import_id_fields = ("name",)


class DictWidget(Widget):
    """Simple widget that uses JSON to store and load dictionary fields."""

    def clean(self, value, row=None, **kwargs):
        """Returns Python object from export representation."""
        return json.loads(value)

    def render(self, value, obj=None):
        """Converts Python object to export representation."""
        return json.dumps(value)


class AnsibleNamespaceMetadataResource(BaseContentResource):
    """
    Resource for import/export of ansible_namespace-metadata-content entities.
    """

    links = fields.Field(attribute="links", column_name="links", widget=DictWidget())

    def before_import_row(self, row, **kwargs):
        """
        Finds and sets namespace using the name of the namespace metadata.
        """
        super().before_import_row(row, **kwargs)

        namespace = AnsibleNamespace.objects.get(name=row["name"])
        row["namespace"] = str(namespace.pk)
        if row["avatar_sha256"] == "":
            row["avatar_sha256"] = None

    def set_up_queryset(self):
        """
        :return: AnsibleNamespaceMetadata content specific to a specified repo-version.
        """
        return AnsibleNamespaceMetadata.objects.filter(pk__in=self.repo_version.content)

    class Meta:
        model = AnsibleNamespaceMetadata
        import_id_fields = model.natural_key_fields()


class CollectionVersionContentResource(BaseContentResource):
    """
    Resource for import/export of ansible_collectionversion-content entities.
    """

    def before_import_row(self, row, **kwargs):
        """
        Finds and sets collection using name and namespace.

        Args:
            row (tablib.Dataset row): incoming import-row representing a single CollectionVersion.
            kwargs: args passed along from the import() call.
        """
        super().before_import_row(row, **kwargs)

        col = Collection.objects.get(name=row["name"], namespace=row["namespace"])
        row["collection"] = str(col.pk)
        row.pop("is_highest", None)

    def set_up_queryset(self):
        """
        :return: CollectionVersion content specific to a specified repo-version.
        """
        return CollectionVersion.objects.filter(pk__in=self.repo_version.content)

    class Meta:
        model = CollectionVersion
        import_id_fields = model.natural_key_fields()


class CollectionVersionSignatureResource(BaseContentResource):
    """
    Resource for import/export of ansible_collectionversionsignature entities.
    """

    def set_up_queryset(self):
        """
        :return: CollectionVersionSignature content specific to a specified repo-version.
        """
        return CollectionVersionSignature.objects.filter(pk__in=self.repo_version.content)

    def before_import_row(self, row, **kwargs):
        """
        Finds and sets collection version using upstream_id.

        Args:
            row (tablib.Dataset row): incoming import-row representing a single content.
            kwargs: args passed along from the import() call.
        """
        super().before_import_row(row, **kwargs)

        cv = CollectionVersion.objects.get(upstream_id=row["signed_collection"])
        row["signed_collection"] = str(cv.pk)

    class Meta:
        model = CollectionVersionSignature
        import_id_fields = ("pubkey_fingerprint", "signed_collection")
        exclude = BaseContentResource.Meta.exclude + ("signing_service",)


class CollectionVersionMarkResource(BaseContentResource):
    """
    Resource for import/export of ansible_collectionversionmark entities.
    """

    def set_up_queryset(self):
        """
        :return: CollectionVersionMark content specific to a specified repo-version.
        """
        return CollectionVersionMark.objects.filter(pk__in=self.repo_version.content)

    def before_import_row(self, row, **kwargs):
        """
        Finds and sets collection version using upstream_id.

        Args:
            row (tablib.Dataset row): incoming import-row representing a single content.
            kwargs: args passed along from the import() call.
        """
        super().before_import_row(row, **kwargs)

        cv = CollectionVersion.objects.get(upstream_id=row["marked_collection"])
        row["marked_collection"] = str(cv.pk)

    class Meta:
        model = CollectionVersionMark
        import_id_fields = ("value", "marked_collection")


class CollectionResource(QueryModelResource):
    """
    Resource for import/export of ansible_collection entities.
    """

    def set_up_queryset(self):
        """
        :return: Collections specific to a specified repo-version.
        """
        collection_versions = CollectionVersion.objects.filter(pk__in=self.repo_version.content)
        return Collection.objects.filter(
            pk__in=collection_versions.values_list("collection", flat=True)
        )

    class Meta:
        model = Collection
        import_id_fields = ("namespace", "name")


class CollectionDeprecationResource(BaseContentResource):
    """
    Resource for import/export of collection_deprecation-content entities.
    """

    def set_up_queryset(self):
        """
        :return: AnsibleCollectionDeprecated content specific to a specified repo-version.
        """
        return AnsibleCollectionDeprecated.objects.filter(pk__in=self.repo_version.content)

    class Meta:
        model = AnsibleCollectionDeprecated
        import_id_fields = ("namespace", "name")


IMPORT_ORDER = [
    AnsibleNamespaceResource,
    AnsibleNamespaceMetadataResource,
    CollectionResource,
    CollectionDeprecationResource,
    CollectionVersionContentResource,
    CollectionVersionMarkResource,
    CollectionVersionSignatureResource,
    RoleContentResource,
]



import asyncio
from collections.abc import Coroutine
import functools
import hashlib
import json
import logging
import tarfile
from collections import defaultdict
from gettext import gettext as _
from operator import attrgetter
from pathlib import Path
from urllib.parse import urljoin
from uuid import uuid4
import tempfile

import yaml
from aiohttp.client_exceptions import ClientError, ClientResponseError
from asgiref.sync import sync_to_async
from django.db import connection, transaction
from django.db.models import Q
from django.db.utils import IntegrityError
from django.utils.dateparse import parse_datetime
from galaxy_importer.collection import CollectionFilename
from galaxy_importer.collection import import_collection as process_collection
from galaxy_importer.collection import sync_collection
from galaxy_importer.exceptions import ImporterError
from git import GitCommandError, Repo
from pulpcore.plugin.exceptions import DigestValidationError
from pulpcore.plugin.models import (
    Artifact,
    ContentArtifact,
    CreatedResource,
    ProgressReport,
    PulpTemporaryFile,
    Remote,
    RepositoryVersion,
    Task,
)
from pulpcore.plugin.stages import (
    ArtifactDownloader,
    ArtifactSaver,
    ContentSaver,
    ContentAssociation,
    EndStage,
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    GenericDownloader,
    QueryExistingArtifacts,
    QueryExistingContents,
    RemoteArtifactSaver,
    ResolveContentFutures,
    Stage,
    create_pipeline,
)
from pulpcore.plugin.util import get_url, get_domain
from rest_framework.serializers import ValidationError
from semantic_version import SimpleSpec, Version
from semantic_version.base import Always

from pulp_ansible.app.constants import PAGE_SIZE
from pulp_ansible.exceptions import CollectionNotFound
from pulp_ansible.app.models import (
    AnsibleCollectionDeprecated,
    AnsibleNamespace,
    AnsibleNamespaceMetadata,
    AnsibleRepository,
    Collection,
    CollectionImport,
    CollectionRemote,
    CollectionVersion,
    CollectionVersionMark,
    CollectionVersionSignature,
)
from pulp_ansible.app.serializers import CollectionVersionSerializer
from pulp_ansible.app.tasks.utils import (
    RequirementsFileEntry,
    get_file_obj_from_tarball,
    parse_collections_requirements_file,
    parse_metadata,
)
from pulp_ansible.app.utils import set_collection_deferred_fields


try:
    from pulpcore.plugin.serializers import RepositoryVersionSerializer
except ImportError:
    RepositoryVersionSerializer = None


log = logging.getLogger(__name__)

aget_url = sync_to_async(get_url)


def async_cache(f):
    # A simple implementation for caching an async function with hashable arguments.
    _cache = {}

    @functools.wraps(f)
    async def _wrapped_f(*args):
        try:
            return _cache[args]
        except KeyError:
            result = await f(*args)
            _cache[args] = result
            if len(_cache) > 128:
                # This is a bit ad hoc and will give fifo instead of lru behaviour.
                # I doubt it even makes a difference here.
                _cache.pop(next(iter(_cache.keys())))
            return result

    return _wrapped_f


# semantic_version.SimpleSpec interpretes "*" as ">=0.0.0"
class AnsibleSpec(SimpleSpec):
    def __init__(self, expression):
        super().__init__(expression)
        if self.expression == "*":
            self.clause = Always()


@sync_to_async
def _save_collection_version(collection_version, artifact):
    with transaction.atomic():
        collection_version.save()
        ContentArtifact.objects.create(
            artifact=artifact,
            content=collection_version,
            relative_path=collection_version.relative_path,
        )


async def declarative_content_from_git_repo(remote, url, git_ref=None, metadata_only=False):
    """Returns a DeclarativeContent for the Collection in a Git repository."""
    if git_ref:
        try:
            gitrepo = Repo.clone_from(
                url, str(uuid4()), depth=1, branch=git_ref, multi_options=["--recurse-submodules"]
            )
        except GitCommandError:
            gitrepo = Repo.clone_from(url, str(uuid4()), multi_options=["--recurse-submodules"])
            gitrepo.git.checkout(git_ref)
    else:
        gitrepo = Repo.clone_from(
            url, str(uuid4()), depth=1, multi_options=["--recurse-submodules"]
        )
    commit_sha = gitrepo.head.commit.hexsha
    metadata, artifact_path = sync_collection(gitrepo.working_dir, ".")
    artifact = Artifact.init_and_validate(artifact_path)
    if metadata_only:
        metadata["artifact"] = None
        metadata["artifact_url"] = None
    else:
        if existing_artifact := await Artifact.objects.filter(
            sha256=artifact.sha256, pulp_domain=get_domain()
        ).afirst():
            await existing_artifact.atouch()
            artifact = existing_artifact
        else:
            try:
                assert artifact._state.adding is True
                await sync_to_async(artifact.save)()
                assert artifact._state.adding is False
            except IntegrityError:
                artifact = await Artifact.objects.aget(
                    sha256=artifact.sha256, pulp_domain=get_domain()
                )
        metadata["artifact_url"] = await aget_url(artifact)
        metadata["artifact"] = artifact
    metadata["remote_artifact_url"] = "{}/commit/{}".format(url.rstrip("/"), commit_sha)
    metadata["sha256"] = artifact.sha256

    artifact = metadata["artifact"]
    try:
        collection_version = await sync_to_async(create_collection_from_importer)(metadata)
        await _save_collection_version(collection_version, artifact)
    except ValidationError as e:
        if "unique" in str(e):
            namespace = metadata["metadata"]["namespace"]
            name = metadata["metadata"]["name"]
            version = metadata["metadata"]["version"]
        else:
            raise e
        collection_version = await CollectionVersion.objects.aget(
            namespace=namespace, name=name, version=version, pulp_domain=get_domain()
        )
    if artifact is None:
        artifact = Artifact()
    d_artifact = DeclarativeArtifact(
        artifact=artifact,
        url=metadata["remote_artifact_url"],
        relative_path=collection_version.relative_path,
        remote=remote,
        deferred_download=metadata_only,
    )

    # TODO: docs blob support??
    extra_data = {}
    if metadata_only:
        extra_data["d_artifact_files"] = {d_artifact: artifact_path}

    d_content = DeclarativeContent(
        content=collection_version,
        d_artifacts=[d_artifact],
        extra_data=extra_data,
    )
    return d_content


def sync(remote_pk, repository_pk, mirror, optimize):
    """
    Sync Collections with ``remote_pk``, and save a new RepositoryVersion for ``repository_pk``.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.
        optimize (boolean): Whether to optimize sync or not.

    Raises:
        ValueError: If the remote does not specify a URL to sync.

    """
    remote = CollectionRemote.objects.get(pk=remote_pk)
    repository = AnsibleRepository.objects.get(pk=repository_pk)

    is_repo_remote = repository.remote is not None and remote.pk == repository.remote.pk

    if not remote.url:
        raise ValueError(_("A CollectionRemote must have a 'url' specified to synchronize."))

    first_stage = CollectionSyncFirstStage(remote, repository, is_repo_remote, optimize)
    if first_stage.should_sync:
        set_collection_deferred_fields(["docs_blob", "manifest", "files", "contents"])
        d_version = AnsibleDeclarativeVersion(first_stage, repository, mirror=mirror)
        repository_version = d_version.create()

        if repository_version is not None:
            repository.last_synced_metadata_time = first_stage.last_synced_metadata_time
            repository.save(update_fields=["last_synced_metadata_time"])
            if RepositoryVersionSerializer is not None:
                return RepositoryVersionSerializer(
                    instance=repository_version, context={"request": None}
                ).data
    else:
        log.debug(_("no-op: remote wasn't updated since last sync."))


def import_collection(
    temp_file_pk,
    repository_pk=None,
    expected_namespace=None,
    expected_name=None,
    expected_version=None,
):
    """
    Create a Collection from an uploaded artifact and optionally validate its expected metadata.

    This task provides optional validation of the `namespace`, `name`, and `version` metadata
    attributes. If the Artifact fails validation or parsing, the Artifact is deleted and the
    Collection is not created.

    This task performs a CollectionImport object get_or_create() to allow import messages to be
    logged.

    Args:
        temp_file_pk (str): The pk of the PulpTemporaryFile to create the Collection from.

    Keyword Args:
        repository_pk (str): Optional. If specified, a new RepositoryVersion will be created for the
            Repository and any new Collection content associated with it.
        expected_namespace (str): Optional. The namespace is validated against the namespace
            specified in the Collection's metadata. If it does not match a ImporterError is
            raised.
        expected_name (str): Optional. The name is validated against the name specified in the
            Collection's metadata. If it does not match a ImporterError is raised.
        expected_version (str): Optional. The version is validated against the version specified in
            the Collection's metadata. If it does not match a ImporterError is raised.

    Raises:
        ImporterError: If the `expected_namespace`, `expected_name`, or `expected_version` do not
            match the metadata in the tarball.

    """
    CollectionImport.objects.get_or_create(task_id=Task.current().pulp_id)

    temp_file = PulpTemporaryFile.objects.get(pk=temp_file_pk)
    filename = CollectionFilename(expected_namespace, expected_name, expected_version)
    log.info(f"Processing collection from {temp_file.file.name}")
    user_facing_logger = logging.getLogger("pulp_ansible.app.tasks.collections.import_collection")

    try:
        with temp_file.file.open() as artifact_file:
            with tarfile.open(fileobj=artifact_file, mode="r") as tar:
                manifest_data = json.load(
                    get_file_obj_from_tarball(tar, "MANIFEST.json", temp_file.file.name)
                )
                files_data = json.load(
                    get_file_obj_from_tarball(tar, "FILES.json", temp_file.file.name)
                )
            url = _get_backend_storage_url(artifact_file)
            artifact_file.seek(0)
            importer_result = process_collection(
                artifact_file, filename=filename, file_url=url, logger=user_facing_logger
            )
        artifact = Artifact.from_pulp_temporary_file(temp_file)
        temp_file = None
        importer_result["artifact_url"] = get_url(artifact)
        importer_result["sha256"] = artifact.sha256
        collection_version = create_collection_from_importer(importer_result)
        collection_version.manifest = manifest_data
        collection_version.files = files_data
        with transaction.atomic():
            collection_version.save()
            ContentArtifact.objects.create(
                artifact=artifact,
                content=collection_version,
                relative_path=collection_version.relative_path,
            )

    except ImporterError as exc:
        log.info(f"Collection processing was not successful: {exc}")
        if temp_file:
            temp_file.delete()
        raise
    except Exception as exc:
        user_facing_logger.error(f"Collection processing was not successful: {exc}")
        if temp_file:
            temp_file.delete()
        raise

    CreatedResource.objects.create(content_object=collection_version)

    if repository_pk:
        repository = AnsibleRepository.objects.get(pk=repository_pk)
        content_q = CollectionVersion.objects.filter(pk=collection_version.pk)
        with repository.new_version() as new_version:
            new_version.add_content(content_q)
        CreatedResource.objects.create(content_object=repository)


def create_collection_from_importer(importer_result):
    """
    Process results from importer.

    Do not perform any database operations, just return an unsaved CollectionVersion.
    """
    collection_info = importer_result["metadata"]

    # Remove fields not used by this model
    collection_info.pop("license_file")
    collection_info.pop("readme")

    # the importer returns many None values. We need to let the defaults in the model prevail
    for key in ["description", "documentation", "homepage", "issues", "repository"]:
        if collection_info[key] is None:
            collection_info.pop(key)

    collection, created = Collection.objects.get_or_create(
        namespace=collection_info["namespace"], name=collection_info["name"]
    )
    collection_version = CollectionVersion(
        **collection_info,
        collection=collection,
        requires_ansible=importer_result.get("requires_ansible"),
        contents=importer_result["contents"],
        docs_blob=importer_result["docs_blob"],
        sha256=importer_result["sha256"],
    )

    serializer_fields = CollectionVersionSerializer.Meta.fields
    data = {k: v for k, v in collection_version.__dict__.items() if k in serializer_fields}

    serializer = CollectionVersionSerializer(data=data)
    serializer.is_valid(raise_exception=True)

    return collection_version


def rebuild_repository_collection_versions_metadata(
    repository_version_pk, namespace=None, name=None, version=None
):
    """Rebuild metadata for all collection versions in a repo."""
    repov = RepositoryVersion.objects.get(pk=repository_version_pk)

    qs = None
    if namespace or name or version:
        qkwargs = {}
        if namespace:
            qkwargs["namespace"] = namespace
        if name:
            qkwargs["name"] = name
        if version:
            qkwargs["version"] = version
        qs = CollectionVersion.objects.filter(**qkwargs)

    qs = repov.get_content(content_qs=qs)
    with ProgressReport(
        message=_("Rebuild collection version metadata (total)"),
        code="rebuild_metadata.total",
        total=qs.count(),
    ) as ptotal, ProgressReport(
        message=_("Rebuild collection version metadata (failed)"), code="rebuild_metadata.failed"
    ) as pfailed:
        for cv in qs:
            try:
                _rebuild_collection_version_meta(cv)
            except Exception as e:
                pfailed.increment()
                log.exception(e)
            ptotal.increment()


def _rebuild_collection_version_meta(content_object):
    """Rebuild metadata for a single collection version."""

    # Cast to get the CV
    collection_version = content_object.cast()

    # where is the artifact?
    artifact = content_object._artifacts.first()

    # call the importer to re-generate meta
    importer_result = process_collection(
        artifact.file, filename=artifact.file.name, file_url=artifact.file.url, logger=None
    )

    # set the new info and save
    collection_version.requires_ansible = importer_result["requires_ansible"]
    collection_version.docs_blob = importer_result["docs_blob"]
    collection_version.contents = importer_result["contents"]
    collection_version.save()


def _get_backend_storage_url(artifact_file):
    """Get artifact url from pulp backend storage."""
    domain = get_domain()
    if (
        domain.storage_class == "pulpcore.app.models.storage.FileSystem"
        or not domain.redirect_to_object_storage
    ):
        url = None
    elif domain.storage_class == "storages.backends.s3boto3.S3Boto3Storage":
        parameters = {"ResponseContentDisposition": "attachment%3Bfilename=archive.tar.gz"}
        url = artifact_file.storage.url(artifact_file.name, parameters=parameters)
    elif domain.storage_class == "storages.backends.azure_storage.AzureStorage":
        url = artifact_file.storage.url(artifact_file.name)
    else:
        raise NotImplementedError(f"The value {domain.storage_class=} was not expected")
    return url


class AnsibleDeclarativeVersion(DeclarativeVersion):
    """
    Subclassed Declarative version creates a custom pipeline for Ansible sync.
    """

    def pipeline_stages(self, new_version):
        """
        Build a list of stages feeding into the ContentUnitAssociation stage.

        This defines the "architecture" of the entire sync.

        Args:
            new_version (:class:`~pulpcore.plugin.models.RepositoryVersion`): The
                new repository version that is going to be built.

        Returns:
            list: List of :class:`~pulpcore.plugin.stages.Stage` instances

        """
        pipeline = [
            self.first_stage,
            QueryExistingArtifacts(),
            ArtifactDownloader(),
            ArtifactSaver(),
            QueryExistingContents(),
            DocsBlobDownloader(),
            AnsibleContentSaver(new_version),
            RemoteArtifactSaver(),
            ResolveContentFutures(),
        ]

        return pipeline

    def create(self):
        """
        Perform the work. This is the long-blocking call where all syncing occurs.

        Returns: The created RepositoryVersion or None if it represents no change from the latest.
        """
        with tempfile.TemporaryDirectory(dir="."):
            with self.repository.new_version() as new_version:
                deprecation_before_sync = {
                    (namespace, name)
                    for namespace, name in AnsibleCollectionDeprecated.objects.filter(
                        pk__in=self.repository.latest_version().content
                    ).values_list("namespace", "name")
                }
                loop = asyncio.get_event_loop()
                stages = self.pipeline_stages(new_version)
                stages.append(UndeprecateStage(deprecation_before_sync))
                stages.append(ContentAssociation(new_version, self.mirror))
                stages.append(EndStage())
                pipeline = create_pipeline(stages)
                loop.run_until_complete(pipeline)

                if deprecation_before_sync:
                    to_undeprecate = Q()
                    for namespace, name in deprecation_before_sync:
                        to_undeprecate |= Q(namespace=namespace, name=name)
                    new_version.remove_content(
                        AnsibleCollectionDeprecated.objects.filter(
                            to_undeprecate, pulp_domain=get_domain()
                        )
                    )

        return new_version if new_version.complete else None


class UndeprecateStage(Stage):
    def __init__(self, deprecation_before_sync):
        self.deprecation_before_sync = deprecation_before_sync

    async def run(self):
        async for batch in self.batches():
            for d_content in batch:
                if isinstance(d_content.content, AnsibleCollectionDeprecated):
                    key = (d_content.content.namespace, d_content.content.name)
                    self.deprecation_before_sync.discard(key)
                await self.put(d_content)


class CollectionSyncFirstStage(Stage):
    """
    The first stage of a pulp_ansible sync pipeline.
    """

    def __init__(self, remote, repository, is_repo_remote, optimize):
        """
        The first stage of a pulp_ansible sync pipeline.

        Args:
            remote (CollectionRemote): The remote data to be used when syncing
            repository (AnsibleRepository): The repository being syncedself.
            is_repo_remote (bool): True if the remote is the repository's remote.
            optimize (boolean): Whether to optimize sync or not.

        """
        super().__init__()
        self.remote = remote
        self.repository = repository
        self.collection_info = parse_collections_requirements_file(remote.requirements_file)
        self.exclude_info = {}
        self.add_dependents = self.collection_info and self.remote.sync_dependencies
        self.signed_only = self.remote.signed_only
        self.already_synced = set()
        self._unpaginated_collection_metadata = None
        self._unpaginated_collection_version_metadata = None
        self.optimize = optimize
        self.last_synced_metadata_time = None
        self.namespace_shas = {}
        self._unpaginated_namespace_metadata = None

        # Interpret download policy
        self.deferred_download = self.remote.policy != Remote.IMMEDIATE

        # Check if we should sync
        self.should_sync = not is_repo_remote or asyncio.get_event_loop().run_until_complete(
            self._should_we_sync()
        )

    @async_cache
    async def _get_root_api(self, root):
        """
        Returns the root api path and api version.

        Based on https://git.io/JTMxE.
        """
        if root == "https://galaxy.ansible.com" or root == "https://galaxy.ansible.com/":
            root = "https://galaxy.ansible.com/api/"

        downloader = self.remote.get_downloader(url=root)

        try:
            api_data = parse_metadata(await downloader.run())
        except (json.decoder.JSONDecodeError, ClientResponseError):
            if root.endswith("/api/"):
                raise

            root = urljoin(root, "api/")
            downloader = self.remote.get_downloader(url=root)
            api_data = parse_metadata(await downloader.run())

        if "available_versions" not in api_data:
            raise RuntimeError(_("Could not find 'available_versions' at {}").format(root))

        if "v3" in api_data.get("available_versions", {}):
            api_version = 3
        elif "v2" in api_data.get("available_versions", {}):
            api_version = 2
        else:
            raise RuntimeError(_("Unsupported API versions at {}").format(root))

        endpoint = f"{root}v{api_version}"

        return endpoint, api_version

    @async_cache
    async def _get_paginated_collection_api(self, root):
        """
        Returns the collection api path and api version.

        Based on https://git.io/JTMxE.
        """
        endpoint, api_version = await self._get_root_api(root)
        return f"{endpoint}/collections/", api_version

    async def _fetch_collection_version_metadata(
        self, api_version, collection_version_url
    ) -> list[Coroutine]:
        downloader = self.remote.get_downloader(url=collection_version_url)
        metadata = parse_metadata(await downloader.run())
        return await self._add_collection_version(api_version, collection_version_url, metadata)

    async def _add_collection_version(
        self, api_version, collection_version_url, metadata
    ) -> list[Coroutine]:
        """Add CollectionVersion to the sync pipeline."""
        url = metadata["download_url"]
        collection_version = CollectionVersion(
            namespace=metadata["namespace"]["name"],
            name=metadata["collection"]["name"],
            version=metadata["version"],
        )
        cv_unique = attrgetter("namespace", "name", "version")(collection_version)
        fullname, version = f"{cv_unique[0]}.{cv_unique[1]}", cv_unique[2]
        if cv_unique in self.already_synced:
            return []

        # Mark the collection version as being processed
        self.already_synced.add(cv_unique)
        await self.parsing_metadata_progress_bar.aincrement()

        if fullname in self.exclude_info and Version(version) in self.exclude_info[fullname]:
            log.debug(_("{}-{} is in excludes list, skipping").format(fullname, version))
            return []

        info = metadata["metadata"]
        signatures = metadata.get("signatures", [])
        marks = metadata.get("marks", [])

        if self.signed_only and not signatures:
            log.debug(_("{}-{} does not have any signatures, skipping").format(fullname, version))
            return []

        dependencies_coros = []
        if self.add_dependents:
            dependencies = info["dependencies"]
            for full_name, version in dependencies.items():
                namespace, name = full_name.split(".")
                req = (namespace, name, version)
                new_req = RequirementsFileEntry(full_name, version=version, source=None)
                if not any([req in self.already_synced, new_req in self.collection_info]):
                    self.collection_info.append(new_req)
                    dependencies_coros.append(self._fetch_collection_metadata(new_req))

        info.pop("tags")
        for attr_name, attr_value in info.items():
            if attr_value is None or attr_name not in collection_version.__dict__:
                continue
            setattr(collection_version, attr_name, attr_value)

        # Process later syncing of CV Namespace Metadata if present
        namespace_sha = metadata["namespace"].get("metadata_sha256")
        if namespace_sha:
            self.namespace_shas[collection_version.namespace] = namespace_sha

        artifact = metadata["artifact"]
        collection_version.sha256 = artifact["sha256"]
        d_artifact = DeclarativeArtifact(
            artifact=Artifact(sha256=artifact["sha256"], size=artifact["size"]),
            url=url,
            relative_path=collection_version.relative_path,
            remote=self.remote,
            deferred_download=self.deferred_download,
        )

        extra_data = {}
        if api_version != 2:  # V2 never implemented the docs-blob requests
            extra_data["docs_blob_url"] = f"{collection_version_url}docs-blob/"

        d_content = DeclarativeContent(
            content=collection_version,
            d_artifacts=[d_artifact],
            extra_data=extra_data,
        )
        del metadata  # Free up memory
        await self.put(d_content)

        if signatures or marks:
            collection_version = await d_content.resolution()
            for signature in signatures:
                sig = signature["signature"]
                cv_signature = CollectionVersionSignature(
                    signed_collection=collection_version,
                    data=sig,
                    digest=hashlib.sha256(sig.encode("utf-8")).hexdigest(),
                    pubkey_fingerprint=signature["pubkey_fingerprint"],
                )
                await self.put(DeclarativeContent(content=cv_signature))

            for mark_value in marks:
                cv_mark = CollectionVersionMark(
                    marked_collection=collection_version,
                    value=mark_value,
                )
                await self.put(DeclarativeContent(content=cv_mark))

        return dependencies_coros

    async def _add_namespace(self, name, namespace_sha):
        """Adds A Namespace metadata content to the pipeline."""

        try:
            ns = await AnsibleNamespaceMetadata.objects.aget(
                metadata_sha256=namespace_sha, pulp_domain=get_domain()
            )
            await self.put(DeclarativeContent(ns))
            await self.parsing_namespace_progress_bar.aincrement()
            return
        except AnsibleNamespaceMetadata.DoesNotExist:
            pass

        endpoint, api_version = await self._get_root_api(self.remote.url)
        namespace_url = f"{endpoint}/namespaces/{name}"
        downloader = self.remote.get_downloader(
            url=namespace_url, silence_errors_for_response_codes={404}
        )
        try:
            result = await downloader.run()
        except FileNotFoundError:
            pass
        else:
            namespace = parse_metadata(result)
            links = namespace.get("links", None)

            # clean up the galaxy API for pulp
            if links:
                namespace["links"] = {x["name"]: x["url"] for x in links}
            else:
                namespace["links"] = dict()

            for key in ("pulp_href", "groups", "id", "related_fields", "users"):
                namespace.pop(key, None)

            url = namespace.pop("avatar_url", None)

            da = (
                [
                    DeclarativeFailsafeArtifact(
                        Artifact(sha256=namespace.get("avatar_sha256")),
                        url=url,
                        remote=self.remote,
                        relative_path=f"{name}-avatar",
                        deferred_download=False,
                        extra_data={"namespace": name},
                    )
                ]
                if url
                else None
            )

            namespace = AnsibleNamespaceMetadata(**namespace)
            dc = DeclarativeContent(namespace, d_artifacts=da)
            await self.put(dc)
            await self.parsing_namespace_progress_bar.aincrement()
            return

        log.info(f"Failed to find namespace {name}")

    async def _add_collection_version_from_git(self, url, gitref, metadata_only) -> list[Coroutine]:
        d_content = await declarative_content_from_git_repo(
            self.remote, url, gitref, metadata_only=False
        )
        await self.put(d_content)
        return []

    def _collection_versions_list_downloader(
        self, api_version, collection_endpoint, namespace, name, page_num, page_size
    ):
        url_without_get_params = f"{collection_endpoint}{namespace}/{name}/versions/"
        if api_version == 2:
            versions_list_url = f"{url_without_get_params}?page={page_num}&page_size={page_size}"
        else:
            offset = (page_num - 1) * page_size
            versions_list_url = f"{url_without_get_params}?limit={page_size}&offset={offset}"
        return self.remote.get_downloader(url=versions_list_url)

    async def _fetch_paginated_collection_metadata(
        self, name, namespace, requirement, source=None
    ) -> list[Coroutine]:
        root = source or self.remote.url
        collection_endpoint, api_version = await self._get_paginated_collection_api(root)
        collection_url = f"{collection_endpoint}{namespace}/{name}"
        collection_metadata_downloader = self.remote.get_downloader(url=collection_url)
        collection_metadata = parse_metadata(await collection_metadata_downloader.run())

        coros = []
        page_num = 1
        while True:
            versions_list_downloader = self._collection_versions_list_downloader(
                api_version, collection_endpoint, namespace, name, page_num, PAGE_SIZE
            )
            collection_versions_list = parse_metadata(await versions_list_downloader.run())
            if api_version == 2:
                collection_versions = collection_versions_list["results"]
            else:
                collection_versions = collection_versions_list["data"]
            for collection_version in collection_versions:
                if Version(collection_version["version"]) in requirement:
                    version_num = collection_version["version"]
                    collection_version_detail_url = f"{collection_url}/versions/{version_num}/"
                    if collection_metadata["deprecated"]:
                        d_content = DeclarativeContent(
                            content=AnsibleCollectionDeprecated(namespace=namespace, name=name),
                        )
                        await self.put(d_content)
                    coros.append(
                        self._fetch_collection_version_metadata(
                            api_version,
                            collection_version_detail_url,
                        )
                    )
            next_value = self._get_response_next_value(api_version, collection_versions_list)
            if not next_value:
                break
            page_num = page_num + 1

        self.parsing_metadata_progress_bar.total += len(coros)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        return coros

    async def _read_from_downloaded_metadata(self, name, namespace, requirement) -> list[Coroutine]:
        coros = []

        if (
            namespace not in self._unpaginated_collection_metadata
            or name not in self._unpaginated_collection_metadata[namespace]
        ):
            raise CollectionNotFound(namespace, name, self.remote.url)

        if self._unpaginated_collection_metadata[namespace][name]["deprecated"]:
            d_content = DeclarativeContent(
                content=AnsibleCollectionDeprecated(namespace=namespace, name=name),
            )
            await self.put(d_content)

        all_versions_of_collection = self._unpaginated_collection_version_metadata[namespace][name]
        for col_version_metadata in all_versions_of_collection:
            if Version(col_version_metadata["version"]) in requirement:
                if "git_url" in col_version_metadata and col_version_metadata["git_url"]:
                    coros.append(
                        self._add_collection_version_from_git(
                            col_version_metadata["git_url"],
                            col_version_metadata["git_commit_sha"],
                            False,
                        )
                    )
                else:
                    collection_version_url = urljoin(
                        self.remote.url, f"{col_version_metadata['href']}"
                    )
                    coros.append(
                        self._add_collection_version(
                            self._api_version, collection_version_url, col_version_metadata
                        )
                    )
        self.parsing_metadata_progress_bar.total += len(coros)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        return coros

    async def _fetch_collection_metadata(self, requirements_entry) -> list[Coroutine]:
        requirement_version = AnsibleSpec(requirements_entry.version)

        namespace, name = requirements_entry.name.split(".")

        if self._unpaginated_collection_version_metadata and requirements_entry.source is None:
            return await self._read_from_downloaded_metadata(name, namespace, requirement_version)
        else:
            return await self._fetch_paginated_collection_metadata(
                name, namespace, requirement_version, requirements_entry.source
            )

    @staticmethod
    def _get_response_next_value(api_version, response):
        if api_version == 2:
            return response.get("next")
        elif links := response.get("links"):
            return links.get("next")
        else:
            return None

    def _collection_list_downloader(self, api_version, collection_endpoint, page_num, page_size):
        if api_version == 2:
            collection_list_url = f"{collection_endpoint}?page={page_num}&page_size={page_size}"
        else:
            offset = (page_num - 1) * page_size
            collection_list_url = f"{collection_endpoint}?limit={page_size}&offset={offset}"
        return self.remote.get_downloader(url=collection_list_url)

    async def _download_unpaginated_metadata(self):
        root_endpoint, api_version = await self._get_root_api(self.remote.url)
        self._api_version = api_version
        if api_version > 2:
            loop = asyncio.get_event_loop()

            collection_endpoint = f"{root_endpoint}/collections/all/"
            excludes_endpoint = f"{root_endpoint}/excludes/"
            col_downloader = self.remote.get_downloader(
                url=collection_endpoint, silence_errors_for_response_status_codes={404}
            )
            exc_downloader = self.remote.get_downloader(
                url=excludes_endpoint, silence_errors_for_response_status_codes={404}
            )
            tasks = [loop.create_task(col_downloader.run()), loop.create_task(exc_downloader.run())]
            col_results, exc_results = await asyncio.gather(*tasks, return_exceptions=True)

            if not isinstance(exc_results, FileNotFoundError):
                excludes_response = parse_metadata(exc_results)
                if excludes_response:
                    try:
                        excludes_list = parse_collections_requirements_file(excludes_response)
                    except ValidationError:
                        pass
                    else:
                        excludes = {r.name: AnsibleSpec(r.version) for r in excludes_list}
                        self.exclude_info.update(excludes)

            if not isinstance(col_results, FileNotFoundError):
                collection_metadata_list = parse_metadata(col_results)

                self._unpaginated_collection_metadata = defaultdict(dict)
                for collection in collection_metadata_list:
                    namespace = collection["namespace"]
                    name = collection["name"]
                    self._unpaginated_collection_metadata[namespace][name] = collection

                collection_version_endpoint = f"{root_endpoint}/collection_versions/all/"
                downloader = self.remote.get_downloader(url=collection_version_endpoint)
                collection_version_metadata_list = parse_metadata(await downloader.run())

                self._unpaginated_collection_version_metadata = defaultdict(
                    lambda: defaultdict(list)
                )
                for collection_version_metadata in collection_version_metadata_list:
                    namespace = collection_version_metadata["namespace"]["name"]
                    name = collection_version_metadata["name"]
                    self._unpaginated_collection_version_metadata[namespace][name].append(
                        collection_version_metadata
                    )

    async def _find_all_collections_from_unpaginated_data(self) -> list[Coroutine]:
        coros = []

        for collection_namespace_dict in self._unpaginated_collection_metadata.values():
            for collection in collection_namespace_dict.values():
                if collection["deprecated"]:
                    d_content = DeclarativeContent(
                        content=AnsibleCollectionDeprecated(
                            namespace=collection["namespace"], name=collection["name"]
                        ),
                    )
                    await self.put(d_content)

        for collections_in_namespace in self._unpaginated_collection_version_metadata.values():
            for collection_versions in collections_in_namespace.values():
                for collection_version in collection_versions:
                    collection_version_url = urljoin(
                        self.remote.url, f"{collection_version['href']}"
                    )
                    coros.append(
                        self._add_collection_version(
                            self._api_version, collection_version_url, collection_version
                        )
                    )

        self.parsing_metadata_progress_bar.total = len(coros)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        return coros

    async def _find_all_collections(self) -> list[Coroutine]:
        if self._unpaginated_collection_version_metadata:
            return await self._find_all_collections_from_unpaginated_data()

        collection_endpoint, api_version = await self._get_paginated_collection_api(self.remote.url)

        coros = []
        page_num = 1
        while True:
            collection_list_downloader = self._collection_list_downloader(
                api_version, collection_endpoint, page_num, PAGE_SIZE
            )
            collection_list = parse_metadata(await collection_list_downloader.run())

            if api_version == 2:
                collections = collection_list["results"]
            else:
                collections = collection_list["data"]

            for collection in collections:
                if api_version == 2:
                    namespace = collection["namespace"]["name"]
                else:
                    namespace = collection["namespace"]
                name = collection["name"]
                requirements_file = RequirementsFileEntry(
                    name=".".join([namespace, name]),
                    version="*",
                    source=None,
                )
                coros.append(self._fetch_collection_metadata(requirements_file))

            next_value = self._get_response_next_value(api_version, collection_list)
            if not next_value:
                break
            page_num = page_num + 1

        self.parsing_metadata_progress_bar.total = len(coros)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        return coros

    async def _should_we_sync(self):
        """Check last synced metadata time."""
        msg = _("no_change: Checking if remote changed since last sync.")
        async with ProgressReport(message=msg, code="sync.no_change") as noop:
            root, api_version = await self._get_root_api(self.remote.url)
            if api_version == 3:
                downloader = self.remote.get_downloader(
                    url=root, silence_errors_for_response_status_codes={404}
                )
                try:
                    metadata = parse_metadata(await downloader.run())
                except FileNotFoundError:
                    return True

                try:
                    self.last_synced_metadata_time = parse_datetime(metadata["published"])
                except KeyError:
                    return True

                if not self.optimize:
                    return True

                sources = set()
                if self.collection_info:
                    sources = {r.source for r in self.collection_info if r.source}
                sources.add(self.remote.url)
                if len(sources) > 1:
                    return True

                if self.last_synced_metadata_time == self.repository.last_synced_metadata_time:
                    noop.message = _(
                        "no-op: {remote} did not change since last sync - {published}".format(
                            remote=self.remote.url, published=self.last_synced_metadata_time
                        )
                    )
                    return False

        return True

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the ansible metadata.
        """
        tasks = []
        new_tasks = []
        loop = asyncio.get_event_loop()

        msg = _("Parsing CollectionVersion Metadata")
        async with ProgressReport(message=msg, code="sync.parsing.metadata", total=0) as pb:
            self.parsing_metadata_progress_bar = pb
            await self._download_unpaginated_metadata()

            if self.collection_info:
                for requirement_entry in self.collection_info:
                    tasks.append(self._fetch_collection_metadata(requirement_entry))
            else:
                tasks.extend(await self._find_all_collections())
            # Process in chunks to limit memory usage
            while True:
                for i in range(0, len(tasks), 100):
                    results = await asyncio.gather(*tasks[i : i + 100])
                    for sublist in results:
                        new_tasks.extend(sublist)

                if not new_tasks:
                    break
                tasks = new_tasks
                new_tasks = []
            # Ensure PR 'total' is correct before stage finishes
            pb.total = pb.done

        tasks = []
        msg = _("Parsing Namespace Metadata")
        async with ProgressReport(
            message=msg, code="sync.parsing.namespace", total=len(self.namespace_shas)
        ) as pr:
            self.parsing_namespace_progress_bar = pr
            for namespace, namespace_sha in self.namespace_shas.items():
                tasks.append(loop.create_task(self._add_namespace(namespace, namespace_sha)))
            await asyncio.gather(*tasks)
            # Ensuring the total is correct, as some avatar download might fail
            pr.total = pr.done


class DeclarativeFailsafeArtifact(DeclarativeArtifact):
    """
    Special handling for downloading Namespace Avatar Artifacts.
    """

    async def download(self):
        """Allow download to fail, but not stop the sync."""
        artifact_copy = self.artifact
        try:
            return await super().download()
        except (ClientError, DigestValidationError) as e:
            # Reset DA so that future stages can properly handle it
            self.artifact = artifact_copy
            self.deferred_download = True
            name = self.extra_data.get("namespace")
            log.info(f"Failed to download namespace avatar: {name} - {e}, Skipping")
            return None


class DocsBlobDownloader(GenericDownloader):
    """
    Stage for downloading docs_blob.

    Args:
        max_concurrent_content (int): The maximum number of
            :class:`~pulpcore.plugin.stages.DeclarativeContent` instances to handle simultaneously.
            Default is 200.
        args: unused positional arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
        kwargs: unused keyword arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
    """

    PROGRESS_REPORTING_MESSAGE = "Downloading Docs Blob"
    PROGRESS_REPORTING_CODE = "sync.downloading.docs_blob"

    async def _handle_content_unit(self, d_content):
        """Handle one content unit.

        Returns:
            The number of downloads
        """
        downloaded = 0
        content_already_saved = not d_content.content._state.adding
        if not content_already_saved and d_content.d_artifacts:
            remote = d_content.d_artifacts[0].remote
            docs_blob = d_content.extra_data.get("docs_blob_url")
            if docs_blob:
                downloaded = 1
                downloader = remote.get_downloader(
                    url=docs_blob, silence_errors_for_response_status_codes={404}
                )
                download_result = await downloader.run()
                if Path(download_result.path).exists():
                    d_content.extra_data["docs_blob_path"] = download_result.path

        await self.put(d_content)
        return downloaded


class AnsibleContentSaver(ContentSaver):
    """
    A modification of ContentSaver stage that additionally saves Ansible plugin specific items.

    Saves Collection objects related to the CollectionVersion content unit.
    """

    def __init__(self, repository_version, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.repository_version = repository_version

    def _pre_save(self, batch):
        """
        Save a batch of Collection objects.

        Args:
            batch (list of :class:`~pulpcore.plugin.stages.DeclarativeContent`): The batch of
                :class:`~pulpcore.plugin.stages.DeclarativeContent` objects to be saved.
        """
        # Sort the batch by natural key to prevent deadlocks
        # Keep this here until added to Pulpcore
        batch.sort(key=lambda x: "".join(map(str, x.content.natural_key())))
        for d_content in batch:
            if d_content is None:
                continue
            if isinstance(d_content.content, CollectionVersion):
                d_content.extra_data["pre_save_pulp_id"] = d_content.content.pulp_id
                d_content.extra_data["pre_save_adding"] = d_content.content._state.adding
                if d_content.content._state.adding:
                    d_content.content = self._handle_collection_version(d_content)
            elif isinstance(d_content.content, AnsibleNamespaceMetadata):
                name = d_content.content.name
                namespace, created = AnsibleNamespace.objects.get_or_create(name=name)
                d_content.content.namespace = namespace
                if d_content.d_artifacts:
                    da = d_content.d_artifacts[0]
                    # Check to see if avatar failed to download, update metadata if so,
                    # so that the avatar should be attemtped to be downloaded again.
                    if da.deferred_download:
                        d_content.d_artifacts = []
                        d_content.content.avatar_sha256 = None
                        d_content.content.metadata_sha256 = None

    def _post_save(self, batch):
        """
        Update the collection version with the docs_blob and files efficiently
        """
        update_count = 0
        skip_count = 0
        with connection.cursor() as cursor:
            for d_content in batch:
                if d_content and isinstance(d_content.content, CollectionVersion):
                    collection_version = d_content.content
                    pre_adding = d_content.extra_data.get("pre_save_adding")
                    pre_id = d_content.extra_data.get("pre_save_pulp_id")
                    id_match = pre_id == collection_version.pulp_id
                    if not (pre_adding and id_match):
                        skip_count += 1
                        continue
                    files = d_content.extra_data.pop("files_raw")
                    if docs_blob_path := d_content.extra_data.get("docs_blob_path"):
                        with open(docs_blob_path, "r") as docs_blob_file:
                            blob = docs_blob_file.read()
                        sql = (
                            "UPDATE ansible_collectionversion"
                            " SET docs_blob = (%s::jsonb)->'docs_blob',"
                            " files = (%s::jsonb)"
                            " WHERE content_ptr_id = %s"
                        )
                        cursor.execute(
                            sql,
                            [blob, files, str(collection_version.pulp_id)],
                        )
                    else:
                        sql = (
                            "UPDATE ansible_collectionversion"
                            " SET files = (%s::jsonb)"
                            " WHERE content_ptr_id = %s"
                        )
                        cursor.execute(
                            sql,
                            [files, str(collection_version.pulp_id)],
                        )
                    update_count += 1
        log.info(
            f"_post_save: updated={update_count}, skipped={skip_count}, batch_size={len(batch)}"
        )

    def _handle_collection_version(self, d_content):
        collection_version = d_content.content

        collection, created = Collection.objects.get_or_create(
            namespace=d_content.content.namespace, name=d_content.content.name
        )
        collection_version.collection = collection

        d_artifact_files = d_content.extra_data.get("d_artifact_files", {})

        for d_artifact in d_content.d_artifacts:
            artifact = d_artifact.artifact
            # TODO change logic when implementing normal on-demand syncing
            # Special Case for Git sync w/ metadata_only=True
            if artifact_file_name := d_artifact_files.get(d_artifact):
                artifact_file = open(artifact_file_name, mode="rb")
            else:
                artifact_file = artifact.file.open()
            with tarfile.open(fileobj=artifact_file, mode="r") as tar:
                # Defer loading FILES.json to avoid memory usage
                files_fo = get_file_obj_from_tarball(tar, "FILES.json", artifact.file.name)
                d_content.extra_data["files_raw"] = files_fo.read().decode("utf-8")

                runtime_metadata = get_file_obj_from_tarball(
                    tar, "meta/runtime.yml", artifact.file.name, raise_exc=False
                )
                if runtime_metadata:
                    runtime_yaml = yaml.safe_load(runtime_metadata)
                    if runtime_yaml:
                        collection_version.requires_ansible = runtime_yaml.get("requires_ansible")
                manifest_data = json.load(
                    get_file_obj_from_tarball(tar, "MANIFEST.json", artifact.file.name)
                )
                collection_version.manifest = manifest_data
                info = manifest_data["collection_info"]
            artifact_file.close()

            # Remove fields not used by this model
            info.pop("license_file")
            info.pop("readme")

            # Update with the additional data from the Collection
            for attr_name, attr_value in info.items():
                if attr_value is None:
                    continue
                setattr(collection_version, attr_name, attr_value)
            return collection_version



import pytest
from pulpcore.tests.functional.utils import PulpTaskError

from pulpcore.client.pulp_ansible import (
    AnsibleRepositorySyncURL,
)


def _sync_and_assert(
    ansible_bindings,
    remote,
    ansible_repo,
    monitor_task,
):
    body = AnsibleRepositorySyncURL(remote=remote.pulp_href)
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sync(ansible_repo.pulp_href, body).task)

    content_response = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{ansible_repo.versions_href}1/"
    )
    assert content_response.count == 1


@pytest.mark.parallel
def test_sync_through_http_proxy(
    ansible_bindings,
    ansible_repo,
    ansible_collection_remote_factory,
    http_proxy,
    monitor_task,
):
    """
    Test syncing through a http proxy.
    """
    ansible_remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
        proxy_url=http_proxy.proxy_url,
    )

    _sync_and_assert(
        ansible_bindings,
        ansible_remote,
        ansible_repo,
        monitor_task,
    )


@pytest.mark.parallel
def test_sync_through_http_proxy_with_auth(
    ansible_bindings,
    ansible_repo,
    ansible_collection_remote_factory,
    http_proxy_with_auth,
    monitor_task,
):
    """
    Test syncing through a http proxy that requires auth.
    """
    ansible_remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
        proxy_url=http_proxy_with_auth.proxy_url,
        proxy_username=http_proxy_with_auth.username,
        proxy_password=http_proxy_with_auth.password,
    )

    _sync_and_assert(
        ansible_bindings,
        ansible_remote,
        ansible_repo,
        monitor_task,
    )


@pytest.mark.parallel
def test_sync_through_http_proxy_with_auth_but_auth_not_configured(
    ansible_bindings,
    ansible_repo,
    ansible_collection_remote_factory,
    http_proxy_with_auth,
    monitor_task,
):
    """
    Test syncing through a http proxy that requires auth, but auth is not configured.
    """
    ansible_remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
        proxy_url=http_proxy_with_auth.proxy_url,
    )

    try:
        _sync_and_assert(
            ansible_bindings,
            ansible_remote,
            ansible_repo,
            monitor_task,
        )
    except PulpTaskError as exc:
        assert ("[PLP0010]" in exc.task.error["description"]) or (
            "407, message='Proxy Authentication Required'" in exc.task.error["description"]
        )



# Generated by Django 4.2.16 on 2024-11-22 12:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0057_collectionversion_sha256_migrate"),
    ]

    operations = [
        # ---------------------------------------------------------------------
        # In the current timeline of migrations, this step seems superfluous.
        # But in version 0.23.0 we shipped a bad version of 0056.
        # This repairs it.
        migrations.AlterUniqueTogether(
            name="collectionversion",
            unique_together={("namespace", "name", "version")},
        ),
        # ---------------------------------------------------------------------
        migrations.AddConstraint(
            model_name="collectionversion",
            constraint=models.UniqueConstraint(
                condition=models.Q(("sha256__isnull", False)),
                fields=("sha256",),
                name="unique_sha256",
            ),
        ),
    ]



# Generated by Django 2.2.16 on 2020-09-22 20:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0023_alter_requirements_file_field'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='collectionversion',
            name='certification',
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-18 16:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0013_pulp_fields'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='collectionversion',
            name='is_certified',
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='certification',
            field=models.CharField(choices=[('certified', 'certified'), ('not_certified', 'not_certified'), ('needs_review', 'needs_review')], default='needs_review', max_length=13),
        ),
    ]



"""Tests related to sync ansible plugin collection content type."""

import pytest


REQUIREMENTS_FILE = "collections:\n  - testing.k8s_demo_collection"


@pytest.fixture(scope="class")
def galaxy_remote(ansible_collection_remote_factory):
    return ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file=REQUIREMENTS_FILE,
        sync_dependencies=False,
    )


@pytest.fixture(scope="class")
def galaxy_repository(ansible_bindings, monitor_task, ansible_repository_factory, galaxy_remote):
    repository = ansible_repository_factory(remote=galaxy_remote.pulp_href)
    result = ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {})
    monitor_task(result.task)
    return repository


@pytest.fixture(scope="class")
def galaxy_distribution(ansible_distribution_factory, galaxy_repository):
    """
    Provide a distribution that serves the testing.k8s_demo_collection.
    """
    return ansible_distribution_factory(repository=galaxy_repository)


class TestSync:
    def test_synced_collection_is_findable(self, ansible_bindings, galaxy_distribution):
        collection = ansible_bindings.PulpAnsibleApiV3CollectionsApi.read(
            "k8s_demo_collection", "testing", galaxy_distribution.base_path
        )

        assert collection.name == "k8s_demo_collection"
        assert collection.namespace == "testing"

    @pytest.mark.parametrize("requirements_file", (None, REQUIREMENTS_FILE))
    @pytest.mark.parametrize("mirror", (True, False))
    def test_sync_collection_from_pulp(
        self,
        ansible_bindings,
        monitor_task,
        ansible_repository_factory,
        ansible_collection_remote_factory,
        galaxy_distribution,
        requirements_file,
        mirror,
    ):
        remote = ansible_collection_remote_factory(
            url=galaxy_distribution.client_url,
            requirements_file=requirements_file,
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        repository = ansible_repository_factory(remote=remote.pulp_href)
        result = ansible_bindings.RepositoriesAnsibleApi.sync(
            repository.pulp_href, {"mirror": mirror}
        )
        monitor_task(result.task)

        content = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert len(content.results) >= 1

    @pytest.mark.parametrize("mirror", (True, False))
    def test_sync_absent_collection_from_pulp_fails(
        self,
        has_pulp_plugin,
        ansible_bindings,
        monitor_task,
        ansible_repository_factory,
        ansible_collection_remote_factory,
        galaxy_distribution,
        mirror,
    ):
        remote = ansible_collection_remote_factory(
            url=galaxy_distribution.client_url,
            requirements_file="collections:\n  - absent.not_present",
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        repository = ansible_repository_factory(remote=remote.pulp_href)
        result = ansible_bindings.RepositoriesAnsibleApi.sync(
            repository.pulp_href, {"mirror": mirror}
        )
        with pytest.raises(Exception) as exc_info:
            monitor_task(result.task)
        if has_pulp_plugin("core", min="3.102"):
            assert "[PLPAN01]" in exc_info.value.task.error["description"]
        assert "absent.not_present" in exc_info.value.task.error["description"]

    @pytest.mark.parametrize("mirror", (True, False))
    def test_resync_is_noop(
        self,
        ansible_bindings,
        monitor_task,
        ansible_repository_factory,
        ansible_collection_remote_factory,
        galaxy_distribution,
        mirror,
    ):
        remote = ansible_collection_remote_factory(
            url=galaxy_distribution.client_url,
            requirements_file=REQUIREMENTS_FILE,
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        repository = ansible_repository_factory(remote=remote.pulp_href)
        result = ansible_bindings.RepositoriesAnsibleApi.sync(
            repository.pulp_href, {"mirror": mirror}
        )
        monitor_task(result.task)

        # Resync
        result = ansible_bindings.RepositoriesAnsibleApi.sync(
            repository.pulp_href, {"mirror": mirror, "optimize": True}
        )
        task = monitor_task(result.task)
        msg = f"no-op: {galaxy_distribution.client_url} did not change since last sync"
        messages = " ".join([r.message for r in task.progress_reports])
        assert msg in messages

    def test_updating_requirements_file_resets_last_synced(
        self, ansible_bindings, monitor_task, galaxy_repository, galaxy_remote
    ):
        repository = ansible_bindings.RepositoriesAnsibleApi.read(galaxy_repository.pulp_href)
        assert repository.last_synced_metadata_time is not None

        response = ansible_bindings.RemotesCollectionApi.partial_update(
            galaxy_remote.pulp_href, {"requirements_file": "collections:\n  - ansible.posix"}
        )
        monitor_task(response.task)

        repository = ansible_bindings.RepositoriesAnsibleApi.read(galaxy_repository.pulp_href)
        assert repository.last_synced_metadata_time is None
